# -*- coding: utf-8 -*-
#
# Copyright (c) 2016-2017 Kevin Chung
# Copyright (c) 2018 German Mendez Bravo (Kronuz)
# Copyright (c) 2020 Mike Kinney
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
#
'''Mech winrm (windows remote management) class.'''
import logging
import sys


import click
from pypsrp.client import Client


from . import utils
from .mech_instance import MechInstance

LOGGER = logging.getLogger('mech')


@click.group(context_settings=utils.context_settings())
def winrm():
    '''Winrm operations.

    Copy files to a Windows instance.

    Fetch files from a Windows instance.

    Run command or powershell on a Windows instance.
    '''


@winrm.command()
@click.argument('instance', required=False)
@click.pass_context
def config(ctx, instance):
    """
    Show winrm configuration.
    """
    cloud_name = ctx.obj['cloud_name']
    LOGGER.debug('cloud_name:%s instance:%s', cloud_name, instance)

    if cloud_name:
        # Note: All winrm ops are supported.
        utils.cloud_run(cloud_name, ['winrm'])
        return

    if instance:
        # single instance
        instances = [instance]
    else:
        # multiple instances
        instances = utils.instances()

    for an_instance in instances:
        inst = MechInstance(an_instance)
        if inst.created:
            click.echo(inst.winrm_config())
        else:
            click.secho("VM ({}) is not created.".format(an_instance), fg="red")


@winrm.command()
@click.argument('instance', required=True)
@click.option('--command', '-c', required=False, metavar='COMMAND',
              help='Command to run on instance (using command prompt).')
@click.option('--powershell', '-p', required=False, metavar='POWERSHELL',
              help='Powershell to run on instance.')
@click.pass_context
def run(ctx, instance, command, powershell):
    """
    Run command or powershell using winrm

    Notes:
        Example command: 'date /T'
        Example powershell: 'Write-Output hello'

    """
    cloud_name = ctx.obj['cloud_name']
    LOGGER.debug('cloud_name:%s instance:%s command:%s powershell:%s',
                 cloud_name, instance, command, powershell)

    if cloud_name:
        # Note: All winrm ops are supported.
        utils.cloud_run(cloud_name, ['winrm'])
        return

    if (command is None or command == '') and (powershell is None or powershell == ''):
        sys.exit(click.style("Command or Powershell is required", fg="red"))

    inst = MechInstance(instance)

    if inst.created:
        if command:
            utils.winrm_execute_cmd(inst, command)
        if powershell:
            utils.winrm_execute_ps(inst, powershell)

    else:
        click.echo("VM not created.")


@winrm.command()
@click.argument('local', required=True)
@click.argument('remote', required=True)
@click.argument('instance', required=True)
@click.pass_context
def copy(ctx, local, remote, instance):
    """
    Copy local file to remote file on instance.
    """
    cloud_name = ctx.obj['cloud_name']
    LOGGER.debug('cloud_name:%s local:%s remote:%s instance:%s',
                 cloud_name, local, remote, instance)

    if cloud_name:
        # Note: All winrm ops are supported.
        utils.cloud_run(cloud_name, ['winrm'])
        return

    inst = MechInstance(instance)

    if inst.created:
        utils.winrm_copy(inst, local, remote)
        click.echo("Copied")


@winrm.command()
@click.argument('remote', required=True)
@click.argument('local', required=True)
@click.argument('instance', required=True)
@click.pass_context
def fetch(ctx, remote, local, instance):
    """
    Fetch remote file from instance.
    """
    cloud_name = ctx.obj['cloud_name']
    LOGGER.debug('cloud_name:%s remote:%s local:%s instance:%s',
                 cloud_name, remote, local, instance)

    if cloud_name:
        # Note: All winrm ops are supported.
        utils.cloud_run(cloud_name, ['winrm'])
        return

    inst = MechInstance(instance)

    if inst.created:
        utils.suppress_urllib3_errors()
        client = Client(inst.get_ip(), username=inst.user, password=inst.password, ssl=False)
        client.fetch(remote, local)
        click.echo("Fetched")



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016-2017 Kevin Chung
# Copyright (c) 2018 German Mendez Bravo (Kronuz)
# Copyright (c) 2020 Mike Kinney
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
#
"""MechCloudInstance class"""

from __future__ import print_function, absolute_import

import sys
import logging


import click


from . import utils

LOGGER = logging.getLogger('mech')


class MechCloudInstance():
    """Class to hold Mech Cloud instances."""

    def __init__(self, name, clouds=None):
        """Constructor for the mech cloud instance."""
        if not name or name == "":
            raise AttributeError("Must provide a name for the cloud instance.")
        self.name = name
        self.hostname = None
        self.username = None
        self.directory = None

    def read_config(self, name):
        '''Read the configuration.'''
        clouds = utils.load_mechcloudfile(False)
        LOGGER.debug("clouds:%s", clouds)

        if clouds.get(name, None):
            self.name = name
        else:
            sys.exit(click.style("Instance ({}) was not found in the "
                                 "Mechcloudfile".format(name), fg="red"))

        self.hostname = clouds[name].get('hostname')
        self.directory = clouds[name].get('directory')
        self.username = clouds[name].get('username')

    def set_hostname(self, hostname):
        '''Setter for hostname.'''
        if hostname is None or hostname == '':
            sys.exit(click.style("A non-blank hostname is required.", fg="red"))
        self.hostname = hostname

    def set_directory(self, directory):
        '''Setter for directory.'''
        if directory is None or directory == '':
            sys.exit(click.style("A non-blank directory is required.", fg="red"))
        if ' ' in directory:
            sys.exit(click.style("A directory cannot contain spaces.", fg="red"))
        self.directory = directory

    def set_username(self, username):
        '''Setter for username.'''
        if username is None or username == '':
            sys.exit(click.style("A non-blank username is required.", fg="red"))
        self.username = username

    def __repr__(self):
        """Return a representation of a Mech Cloud instance."""
        sep = '\n'
        return ('name:{name}{sep}hostname:{hostname}{sep}directory:{directory}'
                '{sep}username:{username}'.
                format(name=self.name, hostname=self.hostname, directory=self.directory,
                       username=self.username, sep=sep))

    def config(self):
        """Return the configuration for this mech clould instance."""
        entry = {}
        entry['name'] = self.name
        entry['hostname'] = self.hostname
        entry['directory'] = self.directory
        entry['username'] = self.username
        return entry

    def init(self):
        """Initialize the cloud instance.
            - add entry to Mechcloudfile
        """
        click.secho("Writing entry to Mechcloudfile...", fg="blue")
        clouds = utils.load_mechcloudfile(False)
        clouds[self.name] = self.config()
        utils.save_mechcloudfile(clouds)
        # create remote directory, if it does not exist
        # Note: If using ~/mike for the directory, then ~ will not be expanded
        # when surrounded by quotes. For this reason, directory should not have
        # any spaces in it.
        click.secho("Creating directory (if necessary)...", fg="blue")
        self.ssh('if ! [ -d {directory} ]; then mkdir {directory} ; fi'
                 .format(directory=self.directory))
        # create virtualenv, if not already created
        click.secho("Creating python virtual environment (if necessary)...", fg="blue")
        self.ssh('cd {directory}; virtualenv -p python3.7 venv'
                 .format(directory=self.directory))
        # install mikemech into that python virtual environment using pip
        click.secho("Installing mikemech into that python virtual environment...", fg="blue")
        self.ssh('source {}/venv/bin/activate && pip install mikemech'.format(self.directory))
        vmrun_found, _, _ = self.ssh('which vmrun', False)
        if vmrun_found == 0:
            click.secho("VMware vmrun was found.", fg="green")
        else:
            click.secho("VMware vmrun was not found.", fg="yellow")
        vbm_found, _, _ = self.ssh('which VBoxManage', False)
        if vbm_found == 0:
            click.secho("Oracle VBoxManage was found.", fg="green")
        else:
            click.secho("Oracle VBoxManage was not found.", fg="yellow")
        if vmrun_found != 0 and vbm_found != 0:
            click.secho("Neither VMware vmrun nor Oracle VBoxManage was found.", fg="red")
            click.secho("The 'mech' command will not be very useful.", fg="red")
        click.echo("Done.")

    def upgrade(self):
        """Upgrade the cloud instance.
        """
        click.secho("Updating pip on cloud instance:({})...".format(self.name), fg="blue")
        self.ssh('source {}/venv/bin/activate && pip install -U pip'.format(self.directory))
        click.secho("Updating mikemech...", fg="blue")
        self.ssh('source {}/venv/bin/activate && pip install -U mikemech'.format(self.directory))
        click.echo("Done.")

    def ssh(self, command, print_output_on_error=True):
        """Run ssh command using internal variables (like username and hostname) and print output.

           Parameters:
              command(str): command to execute (ex: 'chmod +x /tmp/file')

        """
        return_code, stdout, stderr = utils.ssh_with_username(hostname=self.hostname,
                                                              username=self.username,
                                                              command="'" + command + "'")
        LOGGER.debug('return_code:%d stdout:%s stderr:%s', return_code, stdout, stderr)
        if print_output_on_error:
            if return_code != 0:
                click.secho('Warning: Command did not complete successfully.'
                            '(return_code:{})'.format(return_code), fg="red")
                if stdout:
                    click.echo(stdout)
                if stderr:
                    click.echo(stderr)
        return return_code, stdout, stderr



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016-2017 Kevin Chung
# Copyright (c) 2018 German Mendez Bravo (Kronuz)
# Copyright (c) 2020 Mike Kinney
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
#

"""Mech utility functions."""

from __future__ import division, absolute_import

import os
import re
import random
import string
import sys
import json
import tarfile
import fnmatch
import logging
import tempfile
import subprocess
import collections
from shutil import copyfile, rmtree

import requests
import click
from pypsrp.client import Client

from .vmrun import VMrun
import mech.vbm
from .mech_cloud_instance import MechCloudInstance

LOGGER = logging.getLogger('mech')


def main_dir():
    """Return the main directory."""
    return os.getcwd()


def mech_dir():
    """Return the mech directory."""
    return os.path.join(main_dir(), '.mech')


def makedirs(name, mode=0o777):
    """Make directories with mode supplied."""
    try:
        os.makedirs(name, mode)
    except OSError:
        pass


def start_vm(inst):
    """Start VM.
       inst is a MechInstance
    """
    LOGGER.debug('inst:%s', inst)
    started = None
    if inst.provider == 'vmware':
        # Note: user/password is needed for provisioning
        vmrun = VMrun(inst.vmx, user=inst.user, password=inst.password)
        click.secho("Bringing machine ({}) up...".format(inst.name), fg="blue")
        started = vmrun.start(gui=inst.gui)
    else:
        vbm = mech.vbm.VBoxManage()
        if inst.no_nat:
            bridge_adapter = preferred_interface()
            LOGGER.debug("bridge_adapter:%s", bridge_adapter)
            vbm.bridged(inst.name, bridge_adapter=bridge_adapter, quiet=False)
        else:
            vbm.create_hostonly(quiet=True)
            vbm.hostonly(inst.name, quiet=True)
        vbm.start(vmname=inst.name, gui=inst.gui, quiet=True)
        running_vms = vbm.list_running()
        started = None
        if inst.name in running_vms:
            started = True

    if started is None:
        click.secho("VM not started", fg="red")
    else:
        click.secho("Getting IP address...", fg="blue")
        ip_address = inst.get_ip(wait=True)

        if not inst.disable_shared_folders:
            # Note: virtualbox shared folders is before VM is started
            if inst.provider == 'vmware':
                share_folders(inst)
            else:
                # for virtualbox and shared folders, there are two steps:
                # first step is before boot (see above)
                # second step is to create mount point and mount it:
                virtualbox_share_folder_post_boot(inst)

        if ip_address:
            click.secho("VM ({})started on {}".format(inst.name, ip_address), fg="green")
        else:
            click.secho("VM ({}) started on an unknown "
                        "IP address".format(inst.name), fg="green")

        # if not already using preshared key, switch to it
        if not inst.use_psk and inst.auth:
            add_auth(inst)
            inst.switch_to_psk()

        if inst.remove_vagrant:
            del_user(inst, 'vagrant')

        if not inst.disable_provisioning:
            provision(inst, show=False)


def confirm(prompt, default='y'):
    """Confirmation prompt."""
    default = default.lower()
    if default not in ['y', 'n']:
        default = 'y'
    choicebox = '[Y/n]' if default == 'y' else '[y/N]'
    prompt = prompt + ' ' + choicebox + ' '

    while True:
        some_input = input(prompt).strip()
        if some_input == '':
            if default == 'y':
                return True
            else:
                return False

        if re.match('y(?:es)?', some_input, re.IGNORECASE):
            return True

        elif re.match('n(?:o)?', some_input, re.IGNORECASE):
            return False


def unpause_vm(inst):
    """Unpause a VM."""
    if inst.provider == 'vmware':
        vmrun = VMrun(inst.vmx)
        if vmrun.unpause(quiet=True) is not None:
            click.secho("Getting IP address...", fg="blue")
            ip_address = inst.get_ip(wait=True)
            if not inst.disable_shared_folders:
                share_folders(inst)
            else:
                click.secho("Disabling shared folders...", fg="blue")
                vmrun.disable_shared_folders(quiet=False)
            if ip_address:
                click.secho("VM resumed on {}".format(ip_address), fg="green")
            else:
                click.secho("VM resumed on an unknown IP address", fg="green")
        else:
            # Otherwise try starting
            start_vm(inst)
    else:
        vbm = mech.vbm.VBoxManage()
        if vbm.resume(inst.name, quiet=True) is not None:
            click.secho("Getting IP address...", fg="blue")
            ip_address = inst.get_ip(wait=True)
            # Note: disable_shared_folders is not really an option here
            if ip_address:
                click.secho("VM resumed on {}".format(ip_address), fg="green")
            else:
                click.secho("VM resumed on an unknown IP address", fg="green")
        else:
            # Otherwise try starting
            start_vm(inst)


def save_mechfile_entry(mechfile_entry, name, mechfile_should_exist=False):
    """Save the entry to the Mechfile."""
    LOGGER.debug('mechfile_entry:%s name:%s mechfile_should_exist:%s',
                 mechfile_entry, name, mechfile_should_exist)
    mechfile = load_mechfile(mechfile_should_exist)
    mechfile[name] = mechfile_entry
    LOGGER.debug("name:%s mechfile:%s", name, mechfile)
    return save_mechfile(mechfile)


def remove_mechfile_entry(name, mechfile_should_exist=True):
    """Removed the entry from the Mechfile."""
    LOGGER.debug('name:%s mechfile_should_exist:%s', name, mechfile_should_exist)
    mechfile = load_mechfile(mechfile_should_exist)

    if mechfile.get(name):
        del mechfile[name]

    LOGGER.debug("after removing name:%s mechfile:%s", name, mechfile)
    return save_mechfile(mechfile)


def save_mechfile(mechfile):
    """Save the mechfile object (which is a dict) to a file called 'Mechfile'.
       Return True if save was successful.
    """
    LOGGER.debug('mechfile:%s', mechfile)
    with open(os.path.join(main_dir(), 'Mechfile'), 'w+') as the_file:
        json.dump(mechfile, the_file, sort_keys=True, indent=2, separators=(',', ': '))
    return True


def locate(path, glob):
    """Locate a file in the path provided."""
    for root, _, filenames in os.walk(path):
        for filename in filenames:
            if fnmatch.fnmatch(filename, glob):
                return os.path.abspath(os.path.join(root, filename))


def parse_vmx(path):
    """Parse the virtual machine configuration (.vmx) file and return an
       ordered dictionary.
    """
    vmx = collections.OrderedDict()
    with open(path) as the_file:
        for line in the_file:
            line = line.strip().split('=', 1)
            if len(line) > 1:
                vmx[line[0].rstrip()] = line[1].lstrip()
    return vmx


def update_vmx(path, numvcpus=None, memsize=None, no_nat=False):
    """Update the virtual machine configuration (.vmx)
       file with desired settings.
    """
    updated = False

    vmx = parse_vmx(path)

    # Check if there is an existing interface
    has_network = False
    for vmx_key in vmx:
        if vmx_key.startswith('ethernet'):
            has_network = True

    # Write one if there is not
    if not has_network:
        vmx["ethernet0.addresstype"] = "generated"
        vmx["ethernet0.bsdname"] = "en0"
        if not no_nat:
            vmx["ethernet0.connectiontype"] = "nat"
        vmx["ethernet0.displayname"] = "Ethernet"
        vmx["ethernet0.linkstatepropagation.enable"] = "FALSE"
        vmx["ethernet0.pcislotnumber"] = "32"
        vmx["ethernet0.present"] = "TRUE"
        vmx["ethernet0.virtualdev"] = "e1000"
        vmx["ethernet0.wakeonpcktrcv"] = "FALSE"
        click.secho("Added network interface to vmx file", fg="yellow")
        updated = True

    # write out vmx file if memsize or numvcpus was specified
    if numvcpus is not None:
        vmx["numvcpus"] = '"{}"'.format(numvcpus)
        updated = True

    if memsize is not None:
        vmx["memsize"] = '"{}"'.format(memsize)
        updated = True

    if updated:
        with open(path, 'w') as new_vmx:
            for key in vmx:
                value = vmx[key]
                row = "{} = {}".format(key, value)
                new_vmx.write(row + os.linesep)


def load_mechfile(should_exist=True):
    """Load the Mechfile from disk and return the mechfile as a dictionary."""
    mechfile_fullpath = os.path.join(main_dir(), 'Mechfile')
    LOGGER.debug("mechfile_fullpath:%s", mechfile_fullpath)
    if os.path.isfile(mechfile_fullpath):
        with open(mechfile_fullpath) as the_file:
            try:
                mechfile = json.loads(the_file.read())
                LOGGER.debug('mechfile:%s', mechfile)
                return mechfile
            except ValueError:
                click.secho("Invalid Mechfile.", fg="red")
                return {}
    else:
        if should_exist:
            sys.exit(
                click.style(
                    "Could not find a Mechfile in the current directory. \n"
                    "A Mech environment is required to run this command. Run `mech init` \n"
                    "to create a new Mech environment. Or specify the name of the VM you would \n"
                    "like to start with `mech up <name>`. A final option is to change to a \n"
                    "directory with a Mechfile and to try again.", fg="red"))
        else:
            return {}


def default_shared_folders():
    """Return the default shared folders config.
    """
    return [{'share_name': 'mech', 'host_path': '.'}]


def build_mechfile_entry(location, box=None, name=None, box_version=None,
                         shared_folders=None, provider=None, windows=None):
    """Build the Mechfile from the inputs."""
    LOGGER.debug("location:%s name:%s box:%s box_version:%s provider:%s windows:%s",
                 location, name, box, box_version, provider, windows)
    mechfile_entry = {}

    if location is None:
        return mechfile_entry

    mechfile_entry['name'] = name
    mechfile_entry['box'] = box
    mechfile_entry['box_version'] = box_version
    mechfile_entry['provider'] = provider
    if windows is None:
        windows = False
    mechfile_entry['windows'] = windows

    if shared_folders is None:
        shared_folders = default_shared_folders()
    mechfile_entry['shared_folders'] = shared_folders

    if any(location.startswith(s) for s in ('https://', 'http://', 'ftp://')):
        if not name:
            name = 'first'
        mechfile_entry['url'] = location
        return mechfile_entry

    elif location.startswith('file:') or os.path.isfile(re.sub(r'^file:(?://)?', '', location)):
        if not name:
            name = 'first'
        location = re.sub(r'^file:(?://)?', '', location)
        LOGGER.debug('location:%s', location)
        mechfile_entry['file'] = location
        try:
            # see if the location/file is a json file
            with open(location) as the_file:
                # if an exception is not thrown, then set values and continue
                # to the end of the function
                catalog = json.loads(the_file.read())
                LOGGER.debug('catalog:%s', catalog)
        except (json.decoder.JSONDecodeError, ValueError):
            # this means the location/file is probably a .box file
            # or the json is invalid
            return mechfile_entry
        except IOError:
            # cannot open file
            sys.exit(click.style('Error: Cannot open file:({})'.format(location), fg="red"))
    else:
        try:
            account, box, ver = (location.split('/', 2) + ['', ''])[:3]
            if not account or not box:
                sys.exit(click.style("Provided box name is not valid", fg="red"))
            if ver:
                box_version = ver
            click.secho("Loading metadata for box '{}'{}".format(
                location, " ({})".format(box_version) if box_version else ""), fg="blue")
            url = 'https://app.vagrantup.com/{}/boxes/{}'.format(account, box)
            response = requests.get(url)
            response.raise_for_status()
            catalog = response.json()
        except (requests.HTTPError, ValueError) as exc:
            sys.exit(click.style("Bad response from HashiCorp's Vagrant "
                                 "Cloud API: %s" % exc), fg="red")
        except requests.ConnectionError:
            sys.exit(click.style("Couldn't connect to HashiCorp's Vagrant Cloud API", fg="red"))

    LOGGER.debug("catalog:%s name:%s box_version:%s", catalog, name, box_version)
    return catalog_to_mechfile(catalog, name=name, box=box,
                               box_version=box_version, provider=provider)


def catalog_to_mechfile(catalog, name=None, box=None, box_version=None, provider=None):
    """Convert the Hashicorp cloud catalog entry to Mechfile entry."""
    LOGGER.debug('catalog:%s name:%s box:%s box_version:%s', catalog, name, box, box_version)
    if provider is None:
        provider = 'vmware'
    mechfile = {}
    versions = catalog.get('versions', [])
    for ver in versions:
        current_version = ver['version']
        if not box_version or current_version == box_version:
            for a_provider in ver['providers']:
                if provider in a_provider['name']:
                    mechfile['name'] = name
                    mechfile['provider'] = provider
                    mechfile['box'] = catalog['name']
                    mechfile['box_version'] = current_version
                    mechfile['url'] = a_provider['url']
                    mechfile['shared_folders'] = default_shared_folders()
                    return mechfile
    sys.exit(click.style("Couldn't find a compatible VM using "
                         "catalog:{}".format(catalog), fg="red"))


def tar_cmd(*args, **kwargs):
    """Build the tar command to be used to extract the box."""
    try:
        startupinfo = None
        if os.name == "nt":
            startupinfo = subprocess.STARTUPINFO()
            startupinfo.dwFlags |= subprocess.SW_HIDE | subprocess.STARTF_USESHOWWINDOW
        # run the 'tar --help' command to see what capabilities are available
        proc = subprocess.Popen(['tar', '--help'], stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE, startupinfo=startupinfo, text=True)
    except OSError:
        return None
    if proc.returncode:
        return None
    stdoutdata, _ = proc.communicate()
    tar = ['tar']
    # if the tar_cmd has the option enabled *and* the capability was in the 'tar --help'
    # then append that option to the tar command (which is a list of strings)
    if kwargs.get('wildcards') and re.search(r'--wildcards\b', stdoutdata):
        tar.append('--wildcards')
    if kwargs.get('force_local') and re.search(r'--force-local\b', stdoutdata):
        tar.append('--force-local')
    if kwargs.get('fast_read') and sys.platform.startswith('darwin'):
        tar.append('--fast-read')
    tar.extend(args)
    return tar


def init_box(name, box=None, box_version=None, location=None, force=False, save=True,
             instance_path=None, numvcpus=None, memsize=None, no_nat=False, provider=None,
             windows=None):
    """Initialize the box. This includes uncompressing the files
       from the box file and updating the vmx file with
       desired settings (if vmware).

       Return the full path to the vmx or vbox file.

       VMware will just use the files as extracted.
       VirtualBox needs to "import" the ovf. It creates a .vbox file.
    """
    LOGGER.debug("name:%s box:%s box_version:%s location:%s provider:%s windows:%s",
                 name, box, box_version, location, provider, windows)
    if provider is None:
        provider = 'vmware'

    look_for = '*.vmx'
    vbox_path = None
    instance_path_save = instance_path
    if provider == 'virtualbox':
        look_for = '*.ovf'
        vbox_path = locate(instance_path, '*.vbox')
        # need to extract the files somewhere, then we will
        # remove them after importing to virtualbox
        if vbox_path != '':
            instance_path += '_tmp'

    # if we do not find the vmx file nor is the already imported files in place
    found_vmx_or_ovf = locate(instance_path, look_for)
    if not found_vmx_or_ovf and vbox_path != '':
        name_version_box = add_box(
            name=name,
            box=box,
            box_version=box_version,
            location=location,
            force=force,
            save=save,
            provider=provider,
            windows=windows)
        if not name_version_box:
            sys.exit(click.style("Cannot find a valid box with a VMX/OVF "
                                 "file in boxfile", fg="red"))

        box_parts = box.split('/')
        box_dir = os.path.join(*filter(None, (mech_dir(), 'boxes', provider,
                                              box_parts[0], box_parts[1], box_version)))
        box_file = locate(box_dir, '*.box')

        click.secho("Extracting box '{}'...".format(box_file), fg="blue")
        makedirs(instance_path)
        if sys.platform == 'win32':
            cmd = tar_cmd('-xf', box_file, force_local=True)
        else:
            cmd = tar_cmd('-xf', box_file)
        if cmd:
            startupinfo = None
            if os.name == "nt":
                startupinfo = subprocess.STARTUPINFO()
                startupinfo.dwFlags |= subprocess.SW_HIDE | subprocess.STARTF_USESHOWWINDOW
            proc = subprocess.Popen(cmd, cwd=instance_path, startupinfo=startupinfo)
            if proc.wait():
                sys.exit(click.style("Cannot extract box", fg="red"))
        else:
            tar = tarfile.open(box_file, 'r')
            tar.extractall(instance_path)

        if not save and box.startswith(tempfile.gettempdir()):
            os.unlink(box)

    if provider == 'vmware':
        vmx_path = locate(instance_path, '*.vmx')
        if not vmx_path:
            sys.exit(click.style("Cannot locate a VMX file", fg="red"))
        update_vmx(vmx_path, numvcpus=numvcpus, memsize=memsize, no_nat=no_nat)
        return vmx_path
    else:
        ovf_path = locate(instance_path, '*.ovf')
        if not ovf_path:
            sys.exit(click.style("Cannot locate an OVF file", fg="red"))
        LOGGER.debug('ovf_path:%s', ovf_path)
        vbm = mech.vbm.VBoxManage()
        import_results = vbm.importvm(path_to_ovf=ovf_path, name=name,
                                      base_folder=mech_dir(), quiet=True)
        LOGGER.debug('import_results:%s', import_results)
        vbox_path = locate(instance_path_save, '*.vbox')
        if not vbox_path:
            sys.exit(click.style("Cannot locate a vbox file", fg="red"))
        # remove the extracted files
        rmtree(instance_path)
        return vbox_path


def add_box(name=None, box=None, box_version=None, location=None,
            force=False, save=True, provider=None, windows=None):
    """Add a box."""
    # build the dict
    LOGGER.debug('name:%s box:%s box_version:%s location:%s provider:%s windows:%s',
                 name, box, box_version, location, provider, windows)

    mechfile_entry = build_mechfile_entry(
        box=box,
        name=name,
        location=location,
        box_version=box_version,
        provider=provider,
        windows=windows)

    return add_mechfile(
        mechfile_entry,
        name=name,
        box=box,
        location=location,
        box_version=box_version,
        force=force,
        save=save,
        provider=provider,
        windows=windows)


def add_mechfile(mechfile_entry, name=None, box=None, box_version=None,
                 location=None, force=False, save=True, provider=None,
                 windows=None):
    """Add a mechfile entry."""
    LOGGER.debug('mechfile_entry:%s name:%s box:%s box_version:%s location:%s '
                 'provider:%s windows:%s', mechfile_entry, name, box,
                 box_version, location, provider, windows)

    box = mechfile_entry.get('box')
    name = mechfile_entry.get('name')
    box_version = mechfile_entry.get('box_version')
    provider = mechfile_entry.get('provider')
    windows = mechfile_entry.get('windows')

    url = mechfile_entry.get('url')
    box_file = mechfile_entry.get('file')

    if box_file:
        return add_box_file(box=box, box_version=box_version, filename=box_file,
                            force=force, save=save, provider=provider,
                            windows=windows)

    if url:
        return add_box_url(name=name, box=box, box_version=box_version,
                           url=url, force=force, save=save,
                           provider=provider, windows=windows)
    click.secho("Could not find a VMWare compatible VM for '{}'{}".format(
        name, " ({})".format(box_version) if box_version else ""), fg="red")


def add_box_url(name, box, box_version, url, force=False, save=True, provider=None, windows=None):
    """Add a box using the URL."""
    LOGGER.debug('name:%s box:%s box_version:%s url:%s provider:%s windows:%s',
                 name, box, box_version, url, provider, windows)
    box_parts = box.split('/')
    first_box_part = box_parts[0]
    second_box_part = ''
    if len(box_parts) > 1:
        second_box_part = box_parts[1]
    if provider is None:
        provider = 'vmware'
    box_dir = os.path.join(*filter(None, (mech_dir(), 'boxes', provider,
                                          first_box_part, second_box_part, box_version)))
    exists = os.path.exists(box_dir)
    if not exists or force:
        if exists:
            click.secho("Attempting to download provider:{} "
                        "box:'{}'...".format(provider, box), fg="blue")
        else:
            click.secho("Provider:{} Box:'{}' could not be found. "
                        "Attempting to download...".format(provider, box), fg="blue")
        try:
            click.secho("URL: {}".format(url), fg="blue")
            response = requests.get(url, stream=True)
            response.raise_for_status()
            length = int(response.headers['content-length'])
            chunk_size = 1024 * 1024
            with click.progressbar(length=length, label="Downloading") as bar:
                the_file = tempfile.NamedTemporaryFile(delete=False)
                try:
                    for chunk in response.iter_content(chunk_size=chunk_size):
                        if chunk:
                            the_file.write(chunk)
                            bar.update(chunk_size)
                    the_file.close()
                    if response.headers.get('content-type') == 'application/json':
                        # Downloaded URL might be a Vagrant catalog if it's json:
                        catalog = json.load(the_file.name)
                        mechfile = catalog_to_mechfile(catalog, name, box, box_version)
                        return add_mechfile(
                            mechfile,
                            name=name,
                            box_version=box_version,
                            force=force,
                            save=save,
                            windows=windows)
                    else:
                        # Otherwise it must be a valid box:
                        return add_box_file(box=box, box_version=box_version,
                                            filename=the_file.name, url=url, force=force,
                                            save=save, provider=provider, windows=windows)
                finally:
                    os.unlink(the_file.name)
        except requests.HTTPError as exc:
            sys.exit(click.style(("Bad response: %s" % exc), fg="red"))
        except requests.ConnectionError:
            sys.exit(click.style(("Couldn't connect to '%s'" % url), fg="red"))
    return name, box_version, box


def add_box_file(box=None, box_version=None, filename=None, url=None,
                 force=False, save=True, provider=None, windows=None):
    """Add a box using a file as the source. Returns box and box_version."""
    click.secho("\nChecking integrity of provider:{} box:'{}' "
                "\nfilename:{}...".format(provider, box, filename), fg="blue")

    valid_endswith = 'vmx'
    look_for = '*.vmx'
    if provider == 'virtualbox':
        valid_endswith = 'ovf'
        look_for = '*.ovf'

    click.secho("looking for:{}...".format(look_for), fg="blue")
    if sys.platform == 'win32':
        cmd = tar_cmd('-tf', filename, look_for, wildcards=True, fast_read=True, force_local=True)
    else:
        cmd = tar_cmd('-tf', filename, look_for, wildcards=True, fast_read=True)

    if cmd:
        startupinfo = None
        if os.name == "nt":
            startupinfo = subprocess.STARTUPINFO()
            startupinfo.dwFlags |= subprocess.SW_HIDE | subprocess.STARTF_USESHOWWINDOW
        proc = subprocess.Popen(cmd, startupinfo=startupinfo)
        valid_tar = not proc.wait()
    else:
        tar = tarfile.open(filename, 'r')
        files = tar.getnames()
        valid_tar = False
        for i in files:
            if i.endswith(valid_endswith):
                valid_tar = True
                break
            if i.startswith('/') or i.startswith('..'):
                sys.exit(click.style("This box is comprised of filenames "
                                     "starting with '/' or '..' \n"
                                     "Exiting for the safety of your files.", fg="red"))

    if valid_tar:
        click.secho("Valid tar", fg="blue")
        if save:
            boxname = os.path.basename(url if url else filename)
            box = os.path.join(*filter(None, (mech_dir(), 'boxes', provider, box,
                                              box_version, boxname)))
            path = os.path.dirname(box)
            makedirs(path)
            if not os.path.exists(box) or force:
                copyfile(filename, box)
        else:
            box = filename
        return box, box_version


def get_info_for_auth(mech_use=False):
    """Get information (username/pub_key) for authentication."""
    username = os.getlogin()
    pub_key = os.path.expanduser('~/.ssh/id_rsa.pub')
    return {'auth': {'username': username, 'pub_key': pub_key, 'mech_use': mech_use}}


def init_mechfile(location=None, box=None, name=None, box_version=None, add_me=None,
                  use_me=None, provider=None, windows=None):
    """Initialize the Mechfile."""
    LOGGER.debug("name:%s box:%s box_version:%s location:%s add_me:%s use_me:%s"
                 " provider:%s windows:%s", name, box, box_version, location,
                 add_me, use_me, provider, windows)
    mechfile_entry = build_mechfile_entry(
        location=location,
        box=box,
        name=name,
        box_version=box_version,
        provider=provider,
        windows=windows)
    if add_me:
        mechfile_entry.update(get_info_for_auth(use_me))
    LOGGER.debug('mechfile_entry:%s', mechfile_entry)
    return save_mechfile_entry(mechfile_entry, name, mechfile_should_exist=False)


def add_to_mechfile(location=None, box=None, name=None, box_version=None, add_me=None,
                    use_me=None, provider=None, windows=None):
    """Add entry to the Mechfile."""
    LOGGER.debug("name:%s box:%s box_version:%s location:%s add_me:%s use_me:%s "
                 "provider:%s windows:%s", name, box, box_version, location,
                 add_me, use_me, provider, windows)
    this_mech_entry = build_mechfile_entry(
        location=location,
        box=box,
        name=name,
        box_version=box_version,
        provider=provider,
        windows=windows)
    if add_me:
        this_mech_entry.update(get_info_for_auth(use_me))
    LOGGER.debug('this_mech_entry:%s', this_mech_entry)
    return save_mechfile_entry(this_mech_entry, name, mechfile_should_exist=False)


def random_string(string_len=15):
    """Generate a random string of fixed length."""
    letters = string.ascii_lowercase
    return ''.join(random.choice(letters) for i in range(string_len))


def add_auth(instance):
    """Add authentication to VM."""

    if not instance:
        sys.exit(click.style("Need to provide an instance to before "
                             "we can add authentication.", fg="red"))

    if instance.provider == 'vmware' and instance.vmx is None:
        sys.exit(click.style("Need to provide vmx before we can add authentication.", fg="red"))

    if instance.provider == 'virtualbox' and instance.vbox is None:
        sys.exit(click.style("Need to provide vbox before we can add authentication.", fg="red"))

    if instance.user is None or instance.user == '':
        sys.exit(click.style("Need to provide user to add authentication.", fg="red"))

    if instance.password is None or instance.password == '':
        sys.exit(click.style("Need to provide password to add authentication.", fg="red"))

    click.secho('Adding auth to instance:{}'.format(instance.name), fg="green")

    vmrun = VMrun(instance.vmx, instance.user, instance.password)
    # cannot run if vmware tools are not installed
    if not vmrun.installed_tools():
        sys.exit(click.style("Cannot add authentication if VMware Tools "
                             "are not installed.", fg="red"))

    if instance.auth:
        username = instance.auth.get('username', None)
        pub_key = instance.auth.get('pub_key', None)
        if username and pub_key:
            pub_key_path = os.path.expanduser(pub_key)
            LOGGER.debug("pub_key_path:%s pub_key:%s", pub_key_path, pub_key)
            with open(pub_key_path, 'r') as the_file:
                pub_key_contents = the_file.read().strip()
            if pub_key_contents:
                # set the password to some random string
                # user should never need it (sudo should not prompt for a
                # password)
                password = random_string()
                cmd = ('sudo useradd -m -s /bin/bash -p "{password}" {username};'
                       'sudo mkdir /home/{username}/.ssh;'
                       'sudo usermod -aG sudo {username};'
                       'echo "{username} ALL=(ALL) NOPASSWD: ALL" | '
                       'sudo tee -a /etc/sudoers;'
                       'echo "{pub_key_contents}" | '
                       'sudo tee -a /home/{username}/.ssh/authorized_keys;'
                       'sudo chmod 700 /home/{username}/.ssh;'
                       'sudo chown {username}:{username} /home/{username}/.ssh;'
                       'sudo chmod 600 /home/{username}/.ssh/authorized_keys;'
                       'sudo chown {username}:{username} /home/{username}/.ssh/authorized_keys'
                       ).format(username=username, pub_key_contents=pub_key_contents,
                                password=password)
                LOGGER.debug('cmd:%s', cmd)
                results = vmrun.run_script_in_guest('/bin/sh', cmd, quiet=True)
                LOGGER.debug('results:%s', results)
                if results is None:
                    click.secho("Did not add auth", fg="red")
                else:
                    click.secho("Added auth.", fg="red")
            else:
                click.secho("Could not read contents of the pub_key"
                            " file:{}".format(pub_key), fg="green")
        else:
            click.secho("Warning: Need a username and pub_key in auth.", fg="blue")
    else:
        click.secho("No auth to add.", fg="blue")


def vm_ready_based_on_state(state):
    """Return True if the state is one where we can communicate with it (scp/ssh, etc.)
    """
    if state in ["started", "powered on", "running", "unpaused"]:
        return True
    return False


def ssh(instance, command, plain=None, extra=None, command_args=None):
    """Run ssh command.

       Parameters:
          instance(MechInstance): a mech instance
          command(str): command to execute (ex: 'chmod +x /tmp/file')
          plain(bool): use user/pass auth
          extra(str): arguments to pass to ssh
          command_args(str): arguments for command

       Returns:
          return_code(int): 0=success
          stdout(str): Output from the command
          stderr(str): Error from the command

       Note: May not really need the tempfile if self.use_psk==True.
             Using the tempfile, there are options to not add host to the known_hosts files
             which is useful, but could be MITM attacks. Not likely locally, but still
             could be an issue.
    """
    LOGGER.debug('command:%s plain:%s extra:%s command_args:%s',
                 command, plain, extra, command_args)
    if instance.created:
        state = instance.get_vm_state()
        if vm_ready_based_on_state(state):
            config_ssh = instance.config_ssh()
            temp_file = tempfile.NamedTemporaryFile(delete=False)
            try:
                temp_file.write(config_ssh_string(config_ssh).encode('utf-8'))
                temp_file.close()

                cmds = ['ssh']
                if not plain:
                    cmds.extend(('-F', temp_file.name))
                if not plain:
                    cmds.append(config_ssh['Host'])
                if extra:
                    cmds.append(extra)
                if command:
                    cmds.extend(('--', command))
                if command_args:
                    cmds.append(command_args)

                LOGGER.debug('cmds:%s', cmds)

                # if running a script
                if command:
                    result = subprocess.run(cmds, capture_output=True)
                    stdout = result.stdout.decode('utf-8').strip()
                    stderr = result.stderr.decode('utf-8').strip()
                    return result.returncode, stdout, stderr
                else:
                    # interactive
                    return subprocess.call(cmds), None, None
            finally:
                os.unlink(temp_file.name)
        else:
            return 1, '', 'VM not ready({})'.format(state)


def scp(instance, src, dst, dst_is_host, extra=None):
    """Run scp command.
       Note: May not really need the tempfile if self.use_psk==True.
             Using the tempfile, there are options to not add host to the known_hosts files
             which is useful, but could be MITM attacks. Not likely locally, but still
             could be an issue.
    """
    if instance.created:
        state = instance.get_vm_state()
        if vm_ready_based_on_state(state):
            config_ssh = instance.config_ssh()
            temp_file = tempfile.NamedTemporaryFile(delete=False)

            try:
                temp_file.write(config_ssh_string(config_ssh).encode())
                temp_file.close()

                cmds = ['scp']
                cmds.extend(('-F', temp_file.name))
                if extra:
                    cmds.extend(extra)

                host = config_ssh['Host']
                dst = '{}:{}'.format(host, dst) if dst_is_host else dst
                src = '{}:{}'.format(host, src) if not dst_is_host else src
                cmds.extend((src, dst))

                LOGGER.debug(
                    " ".join(
                        "'{}'".format(
                            c.replace(
                                "'",
                                "\\'")) if ' ' in c else c for c in cmds))
                result = subprocess.run(cmds, capture_output=True)
                stdout = result.stdout.decode('utf-8').strip()
                stderr = result.stderr.decode('utf-8').strip()
                return result.returncode, stdout, stderr
            finally:
                os.unlink(temp_file.name)
        else:
            return 1, '', 'VM not ready({})'.format(state)


def del_user(instance, username):
    """Delete a user in guest VM."""

    if not instance:
        sys.exit(click.style("Need to provide an instance before "
                             "we can delete user:{}.".format(username), fg="red"))

    if instance.vmx is None:
        sys.exit(click.style("VM must be created.", fg="red"))

    if instance.user is None:
        sys.exit(click.style("A user is required.", fg="red"))

    if username is None or username == '':
        sys.exit(click.style("A username to delete is required.", fg="red"))

    click.secho('Removing username ({}) from '
                'instance:{}...'.format(username, instance.name), fg="green")

    cmd = 'sudo userdel -fr {}'.format(username)
    LOGGER.debug('cmd:%s', cmd)
    ssh(instance=instance, command=cmd)


def provision(instance, show=False):
    """Provision an instance.

    Args:
        instance (MechInstance): an instance
        show (bool): just print the provisioning

    Notes:
        Valid provision types are:
           file: copies files to instances
           shell: executes scripts

    """

    if not instance:
        sys.exit(click.style("Need to provide an instance to provision().", fg="red"))

    if instance.provider == 'vmware' and instance.vmx is None:
        sys.exit(click.style("Need to provide vmx to provision().", fg="red"))

    if instance.provider == 'virtualbox' and instance.vbox is None:
        sys.exit(click.style("Need to provide vbox to provision().", fg="red"))

    if instance.user is None:
        sys.exit(click.style("Need to provide user to provision().", fg="red"))

    click.secho('Provisioning instance:{}'.format(instance.name), fg="green")

    provisioned = 0
    if instance.provision:
        for i, pro in enumerate(instance.provision):
            provision_type = pro.get('type')
            if provision_type == 'file':
                source = pro.get('source')
                destination = pro.get('destination')
                if show:
                    click.secho(" instance.name:{} provision_type:{} source:{} "
                                "destination:{}".format(instance.name, provision_type,
                                                        source, destination), fg="green")
                else:
                    click.secho("File provisioning (source:{} destination:{})".format(
                        source, destination), fg="green")
                    results = provision_file(instance, source, destination)
                    LOGGER.debug('results:%s', results)
                    if results is None:
                        click.secho("Not Provisioned", fg="red")
                        return
                provisioned += 1

            elif provision_type == 'shell':
                inline = pro.get('inline')
                path = pro.get('path')

                args = pro.get('args')
                if not isinstance(args, list):
                    args = [args]
                if show:
                    click.secho(" instance.name:{} provision_type:{} inline:{} path:{} "
                                "args:{}".format(instance.name, provision_type,
                                                 inline, path, args), fg="green")
                else:
                    click.secho("Inline shell provisioining (inline:{} path:{} args:{}".format(
                                inline, path, args), fg="green")
                    if provision_shell(instance, inline, path, args) is None:
                        click.secho("Not Provisioned", fg="red")
                        return
                provisioned += 1

            elif provision_type == 'ps':
                inline = pro.get('inline')
                path = pro.get('path')

                args = pro.get('args')
                if not isinstance(args, list):
                    args = [args]
                if show:
                    click.secho(" instance.name:{} provision_type:{} inline:{} path:{} "
                                "args:{}".format(instance.name, provision_type,
                                                 inline, path, args), fg="green")
                else:
                    click.secho("Inline ps provisioining (inline:{} path:{} args:{}".format(
                                inline, path, args), fg="green")
                    if provision_ps(instance, inline, path) is None:
                        click.secho("Not Provisioned", fg="red")
                        return
                provisioned += 1

            elif provision_type == 'pyinfra':
                path = pro.get('path')

                args = pro.get('args')
                if not isinstance(args, list):
                    args = [args]
                if show:
                    click.secho(" instance.name:{} provision_type:{} path:{} "
                                "args:{}".format(instance.name, provision_type,
                                                 path, args), fg="green")
                else:
                    click.secho("pyinfra provisioining (path:{} args:{}".format(
                                path, args), fg="green")
                    return_code, stdout, stderr = provision_pyinfra(instance, path, args)
                    if return_code is None:
                        click.secho("Not Provisioned", fg="red")
                        return
                    LOGGER.debug('return_code:%d stdout:%s stderr:%s', return_code, stdout, stderr)
                provisioned += 1

            else:
                click.secho("Not Provisioned - unknown provision type ({}) "
                            "(entries:{})".format(provision_type, i), fg="red")
                return
        else:
            click.secho("VM ({}) Provision {} "
                        "entries".format(instance.name, provisioned), fg="green")
    else:
        click.secho("Nothing to provision", fg="blue")


def winrm_copy(instance, local, remote):
    '''Copy file using winrm.'''
    suppress_urllib3_errors()
    client = Client(instance.get_ip(), username=instance.user,
                    password=instance.password, ssl=False)
    client.copy(local, remote)
    click.echo("Copied")
    # return code, stdout, stderr
    return 0, '', ''


def winrm_execute_cmd(instance, command):
    '''Run command prompt command using winrm.

       Returns:
          return_code
          stdout
          stderr
    '''
    LOGGER.debug('instance.name:%s command:%s', instance.name, command)
    suppress_urllib3_errors()
    client = Client(instance.get_ip(), username=instance.user,
                    password=instance.password, ssl=False)
    stdout, stderr, return_code = client.execute_cmd(command)
    LOGGER.debug('command:%s return_code:%d stdout:%s stderr:%s',
                 command, return_code, stdout, stderr)
    if stdout:
        click.echo(stdout.strip())
    if stderr:
        click.echo(stderr)
    return return_code, stdout, stderr


def winrm_execute_ps(instance, powershell, args=None):
    '''Run powershell using winrm.

       Returns:
          return_code
          stdout
          stderr (which will be '')
    '''
    LOGGER.debug('instance.name:%s powershell:%s args:%s', instance.name, powershell, args)
    suppress_urllib3_errors()

    powershell_with_args = powershell
    if args is not None:
        powershell_with_args = '{} {}'.format(powershell, args)

    client = Client(instance.get_ip(), username=instance.user,
                    password=instance.password, ssl=False)
    output, _, had_errors = client.execute_ps(powershell_with_args)
    LOGGER.debug('powershell:%s output:%s had_errors:%s', powershell, output, had_errors)

    return_code = 0
    if had_errors:
        return_code = 1
    if output:
        click.echo(output)
    return return_code, output, ''


def provision_file(instance, source, destination):
    """Provision from file.

    Args:
        instance (MechInstance): instance of the MechInstance class
        source (str): full path of a file to copy
        source (str): full path where the file is to be copied to

    Returns:
        return_code
        stdout
        stderr

    Notes:
       This function copies a file from host to guest.

    """
    click.secho("Copying ({}) to ({})".format(source, destination), fg="blue")
    if instance.windows is True:
        click.secho("Note: Windows instance.", fg="blue")
        return winrm_copy(instance, source, destination)
    else:
        return scp(instance, source, destination, True)


def create_tempfile_in_guest(instance):
    """Create a tempfile in the guest."""
    cmd = 'tmpfile=$(mktemp); echo $tmpfile'
    _, stdout, _ = ssh(instance=instance, command=cmd)
    return stdout


def winrm_create_tempfile_in_guest(instance):
    """Create a tempfile in the guest using wirrm."""
    ps = '$file=New-TemporaryFile; $file.fullname'
    _, stdout, _ = winrm_execute_ps(instance=instance, powershell=ps)
    return stdout


def provision_shell(instance, inline, script_path, args=None):
    """Provision from shell.

       Note: The script must be copied to guest, then run from there.

    Args:
        instance (MechInstance): instance of the MechInstance class
        inline (bool): run the script inline
        script_path (str): path to the script to run
        args (list of str): arguments to the script

    """
    if args is None:
        args = []
    tmp_path = create_tempfile_in_guest(instance)
    LOGGER.debug('inline:%s script_path:%s args:%s tmp_path:%s',
                 inline, script_path, args, tmp_path)
    if tmp_path is None or tmp_path == '':
        click.secho("Warning: Could not create tempfile in guest.", fg="red")
        return

    try:
        if script_path and os.path.isfile(script_path):
            click.secho("Configuring script {}...".format(script_path), fg="blue")
            results = scp(instance, script_path, tmp_path, True)
            if results is None:
                click.secho("Warning: Could not copy file to guest.", fg="red")
                return
        else:
            if script_path:
                if any(script_path.startswith(s) for s in ('https://', 'http://', 'ftp://')):
                    click.secho("Downloading {}...".format(script_path), fg="blue")
                    try:
                        response = requests.get(script_path)
                        response.raise_for_status()
                        inline = response.read()
                    except requests.HTTPError:
                        return
                    except requests.ConnectionError:
                        return
                else:
                    click.secho("Cannot open {}".format(script_path), fg="red")
                    return

            if not inline:
                click.secho("No script to execute", fg="red")
                return

            click.secho("Configuring script to run inline...", fg="blue")
            the_file = tempfile.NamedTemporaryFile(delete=False)
            try:
                the_file.write(str.encode(inline))
                the_file.close()
                scp(instance, the_file.name, tmp_path, True)
            finally:
                os.unlink(the_file.name)

        click.secho("Configuring environment...", fg="blue")
        make_executable = "chmod +x '{}'".format(tmp_path)
        LOGGER.debug('make_executable:%s', make_executable)
        if ssh(instance=instance, command=make_executable) is None:
            click.secho("Warning: Could not configure script in the environment.", fg="red")
            return

        click.secho("Executing program...", fg="blue")
        args_string = ' '.join([str(elem) for elem in args])
        LOGGER.debug('args:%s args_string:%s', args, args_string)
        return ssh(instance=instance, command=tmp_path, command_args=args_string)

    finally:
        return ssh(instance, 'rm -f "{}"'.format(tmp_path))


def provision_ps(instance, inline, script_path):
    """Provision from powershell.

       Notes:
          1) The script will be run from the host, not the guest.
          2) Args are not used at this time.

    Args:
        instance (MechInstance): instance of the MechInstance class
        inline (bool): run the script inline
        script_path (str): path to the script to run
        args (list of str): arguments to the script

    """
    LOGGER.debug('inline:%s script_path:%s', inline, script_path)

    ps = ''

    if script_path and os.path.isfile(script_path):
        # we have a local file with powershell, no need to download
        # just load into a variable
        with open(script_path, 'r') as script_file:
            ps = script_file.read().strip()
    else:
        if inline:
            # we have inline script to run
            ps = inline
        else:
            if script_path:
                if any(script_path.startswith(s) for s in ('https://', 'http://', 'ftp://')):
                    # looks like we need to download the powershell
                    click.secho("Downloading {}...".format(script_path), fg="blue")
                    try:
                        response = requests.get(script_path)
                        response.raise_for_status()
                        ps = response.read()
                    except requests.HTTPError:
                        return
                    except requests.ConnectionError:
                        return
                else:
                    click.secho("Cannot open {}".format(script_path), fg="red")
                    return
            else:
                click.secho("Need to provide inline or script to run powershell", fg="red")
                return

    click.secho("Executing powershell...({})".format(ps), fg="blue")
    return_code, stdout, stderr = winrm_execute_ps(instance=instance,
                                                   powershell=ps)
    return return_code, stdout, stderr


def provision_pyinfra(instance, script_path, args=None):
    """Provision using pyinfra.

    Args:
        instance (MechInstance): instance of the MechInstance class
        script_path (str): path to the script to run, must end with .py
        args (list of str): arguments to the script

    Return:
        return_code(int): return code of the process (0=success)
        stdout(str): standard output
        stderr(str): standard error


    """
    if args is None:
        args = []

    LOGGER.debug('instance.name:%s script_path:%s args:%s', instance.name, script_path, args)

    if script_path and os.path.isfile(script_path):
        return run_pyinfra_script(instance.get_ip(), instance.user,
                                  password=instance.password,
                                  script_path=script_path, args=args)
    else:
        if script_path:
            if any(script_path.startswith(s) for s in ('https://', 'http://', 'ftp://')):
                click.secho("Downloading {}...".format(script_path), fg="blue")
                try:
                    response = requests.get(script_path)
                    response.raise_for_status()
                    pyinfra_remote_contents = response.text
                except requests.HTTPError:
                    return
                except requests.ConnectionError:
                    return
            else:
                click.secho("Cannot open {}".format(script_path), fg="red")
                return

        LOGGER.debug('pyinfra_remote_contents:%s', pyinfra_remote_contents)
        the_file = tempfile.NamedTemporaryFile(delete=False, suffix='.py')
        try:
            the_file.write(str.encode(pyinfra_remote_contents))
            the_file.close()
            return run_pyinfra_script(host=instance.get_ip(), username=instance.user,
                                      password=instance.password,
                                      script_path=the_file.name, args=args)
        finally:
            os.unlink(the_file.name)


def pyinfra_installed():
    """Return True if the utility 'pyinfra' is installed, otherwise return False.
    """
    results = subprocess.run("pyinfra --version", shell=True, capture_output=True)
    return results.returncode == 0


def run_pyinfra_script(host, username, password=None, script_path=None, args=None):
    """Run a pyinfra script on host.

       Parameters
         - host(string): hostname to run the pyinfra script against, can be either
                       a hostname or ip address
         - username(string): username to use for the authentication to the host
         - password(string): password to use for the authentication to the host,
                             if no password, assume pre-shared key
         - script_path(string): the path of of the pyinfra script
         - args(kwargs): arguments to pass to running the script

       Returns:
         - return_code(int): return code from the run example: 0=success
         - stdout(string): output from the run
         - stderr(string): errors from the run

    """
    LOGGER.debug("host:%s username:%s script_path:%s args:%s",
                 host, username, script_path, args)
    if args is None:
        args = []

    if host is None or host == '':
        click.secho("Warning: A host is required for pyinfra provisioning.", fg="red")
        return

    if username is None or username == '':
        click.secho("Warning: A username is required for pyinfra provisioning.", fg="red")
        return

    if script_path is None or script_path == '':
        click.secho("Warning: A script is required for pyinfra provisioning.", fg="red")
        return

    if not os.path.exists(script_path):
        click.secho("Warning: Could not find the pyinfra "
                    "script ({}).".format(script_path), fg="red")
        return

    if not script_path.endswith('.py'):
        click.secho("Warning: A pyinfra provisioning script must end with .py.", fg="red")
        return

    is_pyinfra_installed = pyinfra_installed()
    if not is_pyinfra_installed:
        click.secho("Warning: pyinfra must be installed.", fg="red")
        return

    click.secho("Going to run ({}) using args({}) on host:{} authenticating with username:{}".
                format(script_path, args, host, username), fg="green")

    user_auth = '--user "{}"'.format(username)
    if password is not None:
        user_auth += ' --password "{}"'.format(username, password)
    command = "pyinfra {host} {user_auth} {script}".format(host=host, user_auth=user_auth,
                                                           script=script_path)
    LOGGER.debug('About to run this pyinfra command:%s', command)
    results = subprocess.run(command, shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    stderr = results.stderr.decode('utf-8')
    return results.returncode, stdout, stderr


def config_ssh_string(config_ssh):
    """Build the ssh-config string from a dict holding the keys/values."""
    ssh_config = "Host {}".format(config_ssh.get('Host', '')) + os.linesep
    for key, value in config_ssh.items():
        if key != 'Host':
            ssh_config += "  {} {}".format(key, value) + os.linesep
    return ssh_config


def share_folders(inst):
    """Share folders.
    Args:
        inst (MechInstance): an instance of the MechInstance class (representing a vm)

    """
    if not inst.disable_shared_folders:
        click.secho("Sharing folders...", fg="blue")

        if inst.provider == 'vmware':
            vmrun = VMrun(inst.vmx)
            vmrun.enable_shared_folders(quiet=False)
        else:
            vbm = mech.vbm.VBoxManage()

        for share in inst.shared_folders:
            share_name = share.get('share_name')
            host_path = share.get('host_path')
            if host_path == '.':
                host_path = main_dir()
            absolute_host_path = os.path.abspath(host_path)
            click.secho("share:{} host_path:{} => "
                        "absolute_host_path:{}".format(share_name, host_path,
                                                       absolute_host_path), fg="blue")
            if inst.provider == 'vmware':
                vmrun.add_shared_folder(share_name, host_path, quiet=True)
            else:
                # for virtualbox, the path must be absolute
                vbm.sharedfolder_add(inst.name, share_name, absolute_host_path)


def virtualbox_share_folder_post_boot(inst):
    """For virtualbox, we need to create a mount point and mount the share."""
    for share in inst.shared_folders:
        share_name = share.get('share_name')
        command = ("sudo mkdir -p /mnt/{share_name};"
                   "sudo mount -t vboxsf {share_name}"
                   " /mnt/{share_name}").format(share_name=share_name)
        ssh(instance=inst, command=command)


def get_fallback_executable(command_name='vmrun'):
    """Get a fallback executable for a command line tool."""
    if 'PATH' in os.environ:
        LOGGER.debug("os.environ['PATH']:%s", os.environ['PATH'])
        for path in os.environ['PATH'].split(os.pathsep):
            vmrun = os.path.join(path, command_name)
            if os.path.exists(vmrun):
                return vmrun
            vmrun = os.path.join(path, command_name + '.exe')
            if os.path.exists(vmrun):
                return vmrun
    return None


def valid_provider(provider):
    """Determine if the provider provided is supported and valid.
       Returns True if it is valid or False if it is not valid.
    """
    if provider in ('vmware', 'virtualbox'):
        return True
    else:
        return False


def get_darwin_executable(command_name='vmrun'):
    """Get the full path for the command to run on a mac host."""
    if command_name == 'vmrun':
        full_command = '/Applications/VMware Fusion.app/Contents/Library/vmrun'
    else:
        full_command = '/usr/local/bin/VBoxManage'
    if os.path.exists(full_command):
        return full_command
    return get_fallback_executable()


def get_win32_executable():
    """Get the full path for the 'vmrun' command on a Windows host.
    """
    import winreg
    reg = winreg.ConnectRegistry(None, winreg.HKEY_LOCAL_MACHINE)
    try:
        key = winreg.OpenKey(reg, 'SOFTWARE\\VMware, Inc.\\VMware Workstation')
        try:
            return os.path.join(winreg.QueryValueEx(key, 'InstallPath')[0], 'vmrun.exe')
        finally:
            winreg.CloseKey(key)
    except WindowsError:
        key = winreg.OpenKey(reg, 'SOFTWARE\\WOW6432Node\\VMware, Inc.\\VMware Workstation')
        try:
            return os.path.join(winreg.QueryValueEx(key, 'InstallPath')[0], 'vmrun.exe')
        finally:
            winreg.CloseKey(key)
    finally:
        reg.Close()
    return get_fallback_executable()


def get_provider(vmrun_executable):
    """
    Identifies the right hosttype for vmrun command (ws | fusion | player)
    """

    if sys.platform == 'darwin':
        return 'fusion'

    for provider in ['ws', 'player', 'fusion']:
        # To determine the provider, try
        # running the vmrun command to see which one works.
        try:
            startupinfo = None
            if os.name == "nt":
                startupinfo = subprocess.STARTUPINFO()
                startupinfo.dwFlags |= subprocess.SW_HIDE | subprocess.STARTF_USESHOWWINDOW
            proc = subprocess.Popen([vmrun_executable,
                                     '-T',
                                     provider,
                                     'list'],
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE,
                                    startupinfo=startupinfo)
        except OSError:
            pass

        proc.communicate()
        if proc.returncode == 0:
            return provider


def save_mechcloudfile(inst):
    """Save the inst (which is a dict) to a file called 'Mechcloudfile'.
       Return True if save was successful.
    """
    LOGGER.debug('inst:%s', inst)
    with open(os.path.join(main_dir(), 'Mechcloudfile'), 'w+') as the_file:
        json.dump(inst, the_file, sort_keys=True, indent=2, separators=(',', ': '))
    return True


def remove_mechcloudfile_entry(name, should_exist=True):
    """Remove an entry from the Mechcloudfile."""
    LOGGER.debug('name:%s should_exist:%s', name, should_exist)
    clouds = load_mechcloudfile(should_exist)

    if clouds.get(name):
        del clouds[name]

    LOGGER.debug("after removing name:%s clouds:%s", name, clouds)
    return save_mechcloudfile(clouds)


def load_mechcloudfile(should_exist=True):
    """Load the Mechcloudfile from disk and return as a dictionary."""
    fullpath = os.path.join(main_dir(), 'Mechcloudfile')
    LOGGER.debug("fullpath:%s", fullpath)
    if os.path.isfile(fullpath):
        with open(fullpath) as the_file:
            try:
                clouds = json.loads(the_file.read())
                LOGGER.debug('clouds:%s', clouds)
                return clouds
            except ValueError:
                click.secho("Invalid Mechcloudfile.", fg="red")
                return {}
    else:
        if should_exist:
            sys.exit(
                click.style(
                    "Could not find a Mechcloudfile in the current directory. \n"
                    "A Mech Cloud configuration is required to run this command.\n"
                    "Run `mech cloud init` to create a new Mech Cloud environment.", fg="red"))
        else:
            return {}


def cloud_exists(cloud_name):
    """Return True if the cloud exists in the Mechcloudfile."""
    if cloud_name is None or cloud_name == '':
        return False
    clouds = load_mechcloudfile(False)
    if clouds.get(cloud_name) is not None:
        return True
    else:
        return False


def ssh_with_username(hostname, username, command):
    """Run the command on a host using the username.
    """
    if hostname != '' and username != '' and command != '':
        command = 'ssh {username}@{hostname} -- {command}'.format(username=username,
                                                                  hostname=hostname,
                                                                  command=command)

        LOGGER.debug('command:%s', command)
        result = subprocess.run(command, shell=True, capture_output=True)
        stdout = result.stdout.decode('utf-8')
        stderr = result.stderr.decode('utf-8')
        return result.returncode, stdout, stderr


def report_provider(provider):
    """Check if this provider is available."""
    okay = True
    if not valid_provider(provider):
        click.secho("Invalid provider ({})".format(provider), fg="red")
        okay = False
    if provider == 'vmware':
        vmrun = VMrun()
        if not vmrun.installed():
            click.secho("Warning: Provider is not available.", fg="red")
            click.secho("Install VMware or change provider.", fg="red")
            okay = False
    elif provider == 'virtualbox':
        vbm = mech.vbm.VBoxManage()
        if not vbm.installed():
            click.secho("Warning: Provider is not available.", fg="red")
            click.secho("Install Virtualbox or change provider.", fg="red")
            okay = False
    return okay


def kill_pids(pids):
    """Kill all pids."""
    for pid in pids:
        results = subprocess.run(args='kill {}'.format(pid), shell=True, capture_output=True)
        if results.returncode != 0:
            click.secho("Could not kill pid:{}".format(pid))


def find_pids(search_string):
    """Return all pids that that match the search_string."""
    pids = []
    results = subprocess.run(args="ps -ef | grep '{}' | grep -v grep"
                             .format(search_string), shell=True, capture_output=True)
    if results.returncode == 0:
        # we found a proc
        stdout = results.stdout.decode('utf-8')
        for line in stdout.split('\n'):
            data = line.split()
            if len(data) > 2:
                # add pid to the collection
                pids.append(data[1])
    return pids


def cleanup_dir_and_vms_from_dir(a_dir, names=None, all_vms=False):
    """Kill processes and remove directory and re-create the directory.
       For VMware VMs, look for the .vmx part_of_dir matches the full path.
       For virtualbox look 'VBoxHeadless --comment <name of the instance>'
       Note: Unfortunately, the name is global, so you can only have one
       'first' instance. Make instance name unique for int tests that use
       virtualbox.
       Use a_dir to '' and all_vms=True to kill and remove all.
   """
    if names is None:
        names = ['first']
    # remove from virtualbox
    for name in names:
        subprocess.run(args="VBoxManage unregistervm {}".format(name),
                       shell=True, capture_output=True)
    # kill vmware processes
    kill_pids(find_pids('vmware-vmx.*' + a_dir + '/.mech/'))
    # try to stop, then kill virtualbox processes (if any), then unregister
    for name in names:
        subprocess.run(args="VBoxManage controlvm {} poweroff".format(name),
                       shell=True, capture_output=True)
        kill_pids(find_pids('VBoxHeadless --comment {} '.format(name)))
        subprocess.run(args="VBoxManage unregistervm {} --delete".format(name),
                       shell=True, capture_output=True)
    # clean up the vm files
    rmtree(a_dir, ignore_errors=True)
    # remove any "dead" instances in virtualbox
    results = subprocess.run(args="VBoxManage list vms", shell=True, capture_output=True)
    stdout = results.stdout.decode('utf-8')
    vms = stdout.split('\n')
    LOGGER.debug('vms:%s', vms)
    for line in vms:
        parts = line.split(' ')
        if all_vms:
            if len(parts) > 1:
                vm = parts[1]
                LOGGER.debug('vm:%s', vm)
                # try to stop it first (in case the vm is "locked")
                subprocess.run(args="VBoxManage controlvm {} poweroff".format(vm),
                               shell=True, capture_output=True)
                subprocess.run(args="VBoxManage unregistervm {} --delete".format(vm),
                               shell=True, capture_output=True)
        else:
            if len(parts) > 1 and parts[0] == '"<inaccessible>"':
                vm = parts[1]
                LOGGER.debug('vm:%s', vm)
                # try to stop it first (in case the vm is "locked")
                subprocess.run(args="VBoxManage controlvm {} poweroff".format(vm),
                               shell=True, capture_output=True)
                subprocess.run(args="VBoxManage unregistervm {} --delete".format(vm),
                               shell=True, capture_output=True)
    # re-create the directory to start afresh
    if a_dir != '':
        os.mkdir(a_dir)


def get_interfaces():
    """Get the network interfaces.
       We may have 'ifconfig' or 'ip' installed.
    """
    interfaces = []

    # first try "ifconfig"
    results = subprocess.run(args='ifconfig', shell=True, capture_output=True)
    if results.returncode == 0:
        each_line = results.stdout.decode('utf-8').split('\n')
        for line in each_line:
            print('line:{}'.format(line))
            if len(line) > 0 and line[0] != ' ' and line[0] != '\t':
                parts = line.split()
                if len(parts) > 1:
                    if parts[0].endswith(':'):
                        interfaces.append(parts[0][:-1])
        return interfaces
    else:
        # try "ip"
        results = subprocess.run(args='ip addr', shell=True, capture_output=True)
        if results.returncode == 0:
            each_line = results.stdout.decode('utf-8').split('\n')
            for line in each_line:
                parts = line.split()
                if len(parts) > 2:
                    if parts[1].endswith(':'):
                        interfaces.append(parts[1][:-1])
            return interfaces
    return interfaces


def preferred_interface():
    """Guess the preferred network interface to use as bridge.
       This is a hack.
    """
    interfaces = get_interfaces()
    LOGGER.debug('interfaces:%s', interfaces)
    for interface in interfaces:
        if interface == 'en0':
            return 'en0'
        if interface == 'eno1':
            return 'eno1'
        if interface == 'eth0':
            return 'eth0'
        if interface == 'enp5s0':
            return 'enp5s0'
    return "en0"


def suppress_urllib3_errors():
    """Suppress the urllib3 errors."""
    # Note: Not really sure why we need to do this, but it seems to suppress the output
    # when using the pypsrp client.
    try:
        from urllib3.connectionpool import log
        log.addFilter(SuppressFilter())
    except:  # noqa: E722
        pass


class SuppressFilter(logging.Filter):
    def filter(self, record):
        return 'unparsed data' not in record.getMessage()


def instances():
    """Return list of instances."""
    return list(load_mechfile())


def cloud_instances():
    """Return list of cloud instances."""
    return list(load_mechcloudfile())


def cloud_run(cloud_name, operations):
    """Run the command on the cloud instance.
    """
    if cloud_name and cloud_name != '':
        mci = MechCloudInstance(cloud_name)
        mci.read_config(cloud_name)

        # find out what args were used on the command line
        # any command after the operation will be appended to
        # the command to run on the remote
        args_list = []
        found_operation = False
        LOGGER.debug('sys.argv:%s', sys.argv)
        for arg in sys.argv:
            if arg in operations:
                found_operation = True
            if found_operation:
                args_list.append(arg)
        args_string = ' '.join(args_list)
        LOGGER.debug('cloud_name:%s operations:%s args_list:%s args_string:%s',
                     cloud_name, operations, args_list, args_string)
        command = ('''ssh {username}@{hostname} -- "cd {directory}; '''
                   '''source {directory}/venv/bin/activate && '''
                   '''mech {args_string}"''').format(hostname=mci.hostname,
                                                     directory=mci.directory,
                                                     username=mci.username,
                                                     args_string=args_string)
        LOGGER.debug('command:%s', command)
        result = subprocess.run(command, shell=True, capture_output=True)
        stdout = result.stdout.decode('utf-8')
        stderr = result.stderr.decode('utf-8')
        if stdout:
            click.echo(stdout)
        if stderr:
            click.echo(stderr)
        return result.returncode, stdout, stderr


# for short and long help options
def context_settings():
    return dict(help_option_names=['-h', '--help'])



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016-2017 Kevin Chung
# Copyright (c) 2018 German Mendez Bravo (Kronuz)
# Copyright (c) 2020 Mike Kinney
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the 'Software'), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
#
'''Mech snapshot functionality.'''
import logging
import sys


import click


from . import utils
from .mech_instance import MechInstance
from .vmrun import VMrun

LOGGER = logging.getLogger('mech')


class MechSnapshotAliasedGroup(click.Group):
    '''Enable click command aliases.'''

    def get_command(self, ctx, cmd_name):
        '''get the alias command'''
        try:
            cmd_name = MECH_SNAPSHOT_ALIASES[cmd_name].name
        except KeyError:
            pass
        return super().get_command(ctx, cmd_name)


@click.group(context_settings=utils.context_settings(), cls=MechSnapshotAliasedGroup)
def snapshot():
    '''Snapshot operations.

    Use snapshots to save a point-in-time of the instance.
    '''


@snapshot.command()
@click.argument('name', required=True)
@click.argument('instance', required=True)
@click.pass_context
def delete(ctx, name, instance):
    '''
    Delete a snapshot taken previously with snapshot save.
    '''
    cloud_name = ctx.obj['cloud_name']
    LOGGER.debug('cloud_name:%s name:%s instance:%s',
                 cloud_name, name, instance)

    if cloud_name:
        # Note: All snapshot ops are supported.
        utils.cloud_run(cloud_name, ['snapshot'])
        return

    inst = MechInstance(instance)

    if inst.provider == 'vmware':
        vmrun = VMrun(inst.vmx)
        if vmrun.delete_snapshot(name) is None:
            click.secho('Cannot delete snapshot ({})'.format(name), fg='red')
        else:
            click.secho('Snapshot {} deleted'.format(name), fg='green')
    else:
        click.secho('Not yet implemented on this platform.', fg='red')


@snapshot.command()
@click.argument('instance', required=False)
@click.pass_context
def list(ctx, instance):
    '''
    List all snapshots taken for an instance.
    '''
    cloud_name = ctx.obj['cloud_name']
    LOGGER.debug('cloud_name:%s instance:%s',
                 cloud_name, instance)

    if cloud_name:
        # Note: All snapshot ops are supported.
        utils.cloud_run(cloud_name, ['snapshot'])
        return

    if instance:
        # single instance
        instances = [instance]
    else:
        # multiple instances
        instances = utils.instances()

    for an_instance in instances:
        inst = MechInstance(an_instance)
        click.echo('Snapshots for instance:{}'.format(an_instance))
        if inst.created:
            if inst.provider == 'vmware':
                vmrun = VMrun(inst.vmx)
                click.echo(vmrun.list_snapshots())
            else:
                click.secho('Not yet implemented on this platform.', fg='red')
        else:
            click.secho('Instance ({}) is not created.'.format(an_instance), fg='red')


@snapshot.command()
@click.argument('name', required=True)
@click.argument('instance', required=True)
@click.pass_context
def save(ctx, name, instance):
    '''
    Take a snapshot of the current state of the instance.

    Notes:
        Snapshots are useful for experimenting in a machine for being able
        to rollback quickly.
    '''
    cloud_name = ctx.obj['cloud_name']
    LOGGER.debug('cloud_name:%s name:%s instance:%s', cloud_name, name, instance)

    if cloud_name:
        # Note: All snapshot ops are supported.
        utils.cloud_run(cloud_name, ['snapshot'])
        return

    inst = MechInstance(instance)
    if inst.created:
        if inst.provider == 'vmware':
            vmrun = VMrun(inst.vmx)
            if vmrun.snapshot(name) is None:
                sys.exit(click.style('Warning: Could not take snapshot.', fg='red'))
            else:
                click.secho('Snapshot ({}) on VM ({}) taken'.format(name, instance), fg='green')
        else:
            click.secho('Not yet implemented on this platform.', fg='red')
    else:
        click.secho('Instance ({}) is not created.'.format(instance), fg='red')


MECH_SNAPSHOT_ALIASES = {
    'ls': list,
    'remove': delete,
    'rm': delete,
}



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016-2017 Kevin Chung
# Copyright (c) 2018 German Mendez Bravo (Kronuz)
# Copyright (c) 2020 Mike Kinney
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
#
'''Mech box functionality. (a box file is a tar file containing all of the files
   for a virtual machine)'''
import fnmatch
import logging
import os
import shutil
import sys

import click

from . import utils

LOGGER = logging.getLogger('mech')


class MechBoxAliasedGroup(click.Group):
    '''Enable click command aliases.'''

    def get_command(self, ctx, cmd_name):
        '''get the alias command'''
        try:
            cmd_name = MECH_BOX_ALIASES[cmd_name].name
        except KeyError:
            pass
        return super().get_command(ctx, cmd_name)


@click.group(context_settings=utils.context_settings(), cls=MechBoxAliasedGroup)
def box():
    '''Box operations.

    A box is typically a compressed tar file that contains all files necessary for
    VMware or Virtualbox to import/use.
    '''


@box.command()
@click.argument('location', required=True)
@click.option('--box-version', metavar='VERSION', help='Constrain to specific box version.')
@click.option('-f', '--force', is_flag=True, default=False, help='Overwrite existing Mechfile.')
@click.option('--provider', metavar='PROVIDER', default='vmware',
              help='Provider (`vmware` or `virtualbox`)')
@click.pass_context
def add(ctx, location, box_version, force, provider):
    """
    Add a box to the catalog of available boxes.

    Notes:
        The location can be a:
            URL (ex: 'http://example.com/foo.box'),
            box file (ex: 'file:/mnt/boxen/foo.box'),
            json file (ex: 'file:/tmp/foo.json'), or
            HashiCorp account/box (ex: 'bento/ubuntu-18.04').
    """
    cloud_name = ctx.obj['cloud_name']
    LOGGER.debug('cloud_name:%s location:%s box_version:%s '
                 'force:%s provider:%s', cloud_name,
                 location, box_version, force, provider)

    if cloud_name:
        # Note: All box ops are supported.
        utils.cloud_run(cloud_name, ['box'])
        return

    if not utils.valid_provider(provider):
        sys.exit(click.style('Need to provide valid provider.', fg='red'))

    utils.add_box(name=None, box=None, location=location, box_version=box_version,
                  force=force, provider=provider)


@box.command()
@click.pass_context
def list(ctx):
    """
    List all available boxes in the catalog.
    """

    cloud_name = ctx.obj['cloud_name']
    LOGGER.debug('cloud_name:%s', cloud_name)

    if cloud_name:
        # Note: All box ops are supported.
        utils.cloud_run(cloud_name, ['box'])
        return

    print("{}\t{}\t{}".format(
        'BOX'.rjust(35),
        'VERSION'.rjust(12),
        'PROVIDER'.rjust(12),
    ))
    path = os.path.abspath(os.path.join(utils.mech_dir(), 'boxes'))
    for root, _, filenames in os.walk(path):
        for _ in fnmatch.filter(filenames, '*.box'):
            # Note: The directory has a leading '/', so we need to discard that value
            directory = root.replace(path, '')
            try:
                _, provider, account, a_box, version = directory.split('/', 4)
            except ValueError:
                provider = ''
                _, account, a_box, version = directory.split('/', 3)
            print("{}\t{}\t{}".format(
                "{}/{}".format(account, a_box).rjust(35),
                version.rjust(12),
                provider.rjust(12),
            ))


@box.command()
@click.option('--name', metavar='BOXNAME', required=True,
              help='Box name (ex: `bento/ubuntu-18.04`).')
@click.option('--provider', metavar='PROVIDER', default='vmware',
              help='Provider (`vmware` or `virtualbox`)')
@click.option('--version', metavar='VERSION', required=True, help='Box version.')
@click.pass_context
def remove(ctx, name, provider, version):
    """
    Remove a box that matches the name, provider and version.
    """
    cloud_name = ctx.obj['cloud_name']
    LOGGER.debug('cloud_name:%s name:%s provider:%s version:%s',
                 cloud_name, name, provider, version)

    if cloud_name:
        # Note: All box ops are supported.
        utils.cloud_run(cloud_name, ['box'])
        return

    if not utils.valid_provider(provider):
        sys.exit(click.style("Need to provide valid provider.", fg="red"))

    path = os.path.abspath(os.path.join(utils.mech_dir(), 'boxes', provider, name, version))
    if os.path.exists(path):
        shutil.rmtree(path)
        print("Removed {} {}".format(name, version))
    else:
        print("No boxes were removed.")


MECH_BOX_ALIASES = {
    'delete': remove,
    'ls': list,
    'rm': remove,
}



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016-2017 Kevin Chung
# Copyright (c) 2018 German Mendez Bravo (Kronuz)
# Copyright (c) 2020 Mike Kinney
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
#
'''Mech cli (command line interface).'''
from .mech import cli
from .mech_box import box
from .mech_cloud import cloud
from .mech_snapshot import snapshot
from .mech_winrm import winrm


cli.add_command(box)
cli.add_command(cloud)
cli.add_command(snapshot)
cli.add_command(winrm)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016-2017 Kevin Chung
# Copyright (c) 2018 German Mendez Bravo (Kronuz)
# Copyright (c) 2020 Mike Kinney
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
#
'''Mech cloud functionality.'''
import logging


import click


from . import utils
from .mech_cloud_instance import MechCloudInstance

LOGGER = logging.getLogger('mech')


class MechCloudAliasedGroup(click.Group):
    '''Enable click command aliases.'''

    def get_command(self, ctx, cmd_name):
        '''get the alias command'''
        try:
            cmd_name = MECH_CLOUD_ALIASES[cmd_name].name
        except KeyError:
            pass
        return super().get_command(ctx, cmd_name)


@click.group(context_settings=utils.context_settings(), cls=MechCloudAliasedGroup)
def cloud():
    '''Cloud operations.

    Notes:
        Mech Cloud is an easy way to expose resources for non-local use.

        If you have a spare desktop/laptop:

            1. Install VMware Workstation and/or Oracle VirtualBox, and

            2. Install pre-requisites ("python3.7+" and "virtualenv")

               For Ubuntu:
                 sudo apt-get install python3.7 python3-pip
                 sudo pip3 install virtualenv

               For Fedora:
                 sudo dnf install python37 python3-pip
                 sudo pip3 install virtualenv

        See https://github.com/Fizzadar/pyinfra/tree/master/examples/virtualbox
        to install virtualbox using `pyinfra`.

        Then add that host as a mech "cloud-instance" using the
        "mech cloud init" command.

        This would allow you to spin up/down virtual machines on that
        computer. For instance, if you have a could-instance called "top",
        you could init and start a VM on the remote computer using:
            mech -C top init bento/ubuntu-18.04
            mech -C top up

        The host's directory needs to have a virtual environment setup in
        "venv" and "mech" needs to be installed under that virtual
        environment, such as: "pip install mikemech". The 'mech cloud init'
        takes care of this for you.

        Virtualbox instances are "global". So, you can have only one
        instance named "first".
    '''


@cloud.command()
@click.argument('hostname', required=True, metavar='HOST')
@click.argument('directory', required=True, metavar='DIR')
@click.argument('username', required=True, metavar='USER')
@click.argument('name', required=True, metavar='NAME')
@click.pass_context
def init(ctx, hostname, directory, username, name):
    """
    Initialize Mechcloudfile and add entry.

    The directory will be created with a python virtual environment (venv)
    and "mikemech" will be installed into it.

    Notes:
       - The "mech cloud init" operation can be run again with same cloud-instance.
         (The values would be updated in the Mechcloudfile).

       - The directory should not contain spaces.

       - If you use '~' in the directory value, be sure it is surrounded by
         single quotes because without quotes or if double quotes (") are used,
         then the '~' will be expanded locally and will not use the remote value.
            Good Example: --directory '~/mikemech'
            Bad Example: --directory "~/mikemech"   (do not use)
    """
    cloud_name = ctx.obj['cloud_name']
    LOGGER.debug('cloud_name:%s hostname:%s directory:%s username:%s name:%s',
                 cloud_name, hostname, directory, username, name)

    if cloud_name:
        # Note: All cloud ops are supported.
        utils.cloud_run(cloud_name, ['cloud'])
        return

    inst = MechCloudInstance(name)
    inst.set_hostname(hostname)
    inst.set_directory(directory)
    inst.set_username(username)
    inst.init()


@cloud.command()
@click.argument('name', required=True, metavar='NAME')
@click.pass_context
def remove(ctx, name):
    """
    Remove a Mech Cloud instance
    """
    cloud_name = ctx.obj['cloud_name']
    LOGGER.debug('cloud_name:%s name:%s', cloud_name, name)

    if cloud_name:
        # Note: All cloud ops are supported.
        utils.cloud_run(cloud_name, ['cloud'])
        return

    if utils.cloud_exists(name):
        utils.remove_mechcloudfile_entry(name=name)
        print("Removed ({}) from mech cloud.".format(name))
        print("Be sure to remove any running virtual machines.")
    else:
        print("Cloud ({}) does not exist in the Mechcloudfile.", name)


@cloud.command()
@click.pass_context
def list(ctx):
    """
    List Mech Clouds
    """
    cloud_name = ctx.obj['cloud_name']
    LOGGER.debug('cloud_name:%s', cloud_name)

    if cloud_name:
        # Note: All cloud ops are supported.
        utils.cloud_run(cloud_name, ['cloud'])
        return

    print('=== mech clouds: ===')
    clouds = utils.cloud_instances()
    for a_cloud in clouds:
        mci = MechCloudInstance(a_cloud)
        mci.read_config(a_cloud)
        print(mci)
        print()


@cloud.command()
@click.argument('name', required=False, metavar='NAME')
@click.pass_context
def upgrade(ctx, name):
    """
    Upgrade 'pip' and 'mikemech' on the cloud instances.
    """
    cloud_name = ctx.obj['cloud_name']
    LOGGER.debug('cloud_name:%s name:%s', cloud_name, name)

    if name:
        instances = [name]
    else:
        instances = utils.cloud_instances()

    for a_name in instances:
        mci = MechCloudInstance(a_name)
        mci.read_config(a_name)
        mci.upgrade()


MECH_CLOUD_ALIASES = {
    'delete': remove,
    'ls': list,
    'rm': remove,
}



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016-2017 Kevin Chung
# Copyright (c) 2018 German Mendez Bravo (Kronuz)
# Copyright (c) 2020 Mike Kinney
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
#
'''Mech cli functionality.'''
import logging
import os
import platform
import shutil
import sys


import click

from . import utils
from .mech_instance import MechInstance
from .vmrun import VMrun
from .vbm import VBoxManage
from .__init__ import __version__

LOGGER = logging.getLogger('mech')


class MechAliasedGroup(click.Group):
    '''Enable click command aliases.'''

    def get_command(self, ctx, cmd_name):
        '''get the alias command'''
        try:
            cmd_name = MECH_ALIASES[cmd_name].name
        except KeyError:
            pass
        return super().get_command(ctx, cmd_name)


@click.group(context_settings=utils.context_settings(), cls=MechAliasedGroup)
@click.option('--debug', is_flag=True, default=False)
@click.option('--cloud')
@click.version_option(version=__version__, message='%(prog)s v%(version)s')
@click.pass_context
def cli(ctx, debug, cloud):
    '''Mech is a command line utility for virtual machine automation.

    Create, start, stop, destroy virtual machines (aka instances) with ease.

    VMware Fusion, VMware Workstation, and/or Oracle VirtualBox can be used.

    Mech is similar to Hashicorp's vagrant utility.
    '''
    logger = logging.getLogger()
    handler = logging.StreamHandler(sys.stderr)
    formatter = logging.Formatter('%(filename)s:%(lineno)s %(funcName)s() '
                                  '%(levelname)s: %(message)s')
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    if debug:
        click.echo('Debug is on')
        LOGGER.setLevel(logging.DEBUG)
        LOGGER.debug('cloud:%s', cloud)

    # ensure that ctx.obj exists and is a dict
    ctx.ensure_object(dict)
    ctx.obj['debug'] = debug
    ctx.obj['cloud_name'] = cloud


@cli.command()
@click.option('--detail', '-d', is_flag=True, help='Print detailed info.')
@click.argument('instance', required=False)
@click.pass_context
def list(ctx, detail, instance):
    '''Lists all available instances (using Mechfile)'''

    cloud_name = ctx.obj['cloud_name']
    LOGGER.debug('detail:%s cloud_name:%s', detail, cloud_name)

    if cloud_name:
        utils.cloud_run(cloud_name, ['list', 'ls'])
        return

    if detail:
        click.echo('Instance Details')
        click.echo()
    else:
        click.echo('{}\t{}\t{}\t{}\t{}\t{}'.format(
            'NAME'.rjust(20),
            'ADDRESS'.rjust(15),
            'BOX'.rjust(35),
            'VERSION'.rjust(12),
            'PROVIDER'.rjust(12),
            'STATE'.rjust(12),
        ))

    if instance:
        instances = [instance]
    else:
        instances = utils.instances()

    LOGGER.debug('instances:%s', instances)
    mechfiles = utils.load_mechfile()
    LOGGER.debug('mechfiles:%s', mechfiles)
    for name in instances:
        LOGGER.debug('name:%s', name)
        inst = MechInstance(name, mechfiles)
        vm_state = 'unknown'
        if inst.created:
            ip_address = inst.get_ip()
            vm_state = inst.get_vm_state()
            if vm_state is None:
                vm_state = 'unknown'
            if ip_address is None:
                ip_address = 'poweroff'
            elif not ip_address:
                ip_address = 'running'
        else:
            ip_address = 'notcreated'
            vm_state = 'notcreated'

        if detail:
            click.echo('==================================')
            click.echo('From Mechfile:')
            click.echo(inst)
            click.echo()
            if inst.provider == 'virtualbox':
                if inst.created:
                    click.echo('From virtualbox:')
                    click.echo(inst.get_vm_info())
        else:
            # deal with box_version being none
            box_version = inst.box_version
            if inst.box_version is None:
                box_version = ''
            click.echo('{}\t{}\t{}\t{}\t{}\t{}'.format(
                name.rjust(20),
                ip_address.rjust(15),
                inst.box.rjust(35),
                box_version.rjust(12),
                inst.provider.rjust(12),
                vm_state.rjust(12),
            ))


@cli.command()
@click.pass_context
def support(ctx):
    '''Show support info.'''

    cloud_name = ctx.obj['cloud_name']
    LOGGER.debug('cloud_name:%s', cloud_name)

    if cloud_name:
        utils.cloud_run(cloud_name, ['support'])
        return

    click.echo('If you are having issues with mikemech or wish to make feature requests, ')
    click.echo('please check out the GitHub issues at https://github.com/mkinney/mech/issues .')
    click.echo('')
    click.echo('When adding an issue, be sure to include the following:')
    click.echo(' System: {0}'.format(platform.system()))
    click.echo('   Platform: {0}'.format(platform.platform()))
    click.echo('   Release: {0}'.format(platform.uname().release))
    click.echo('   Machine: {0}'.format(platform.uname().machine))
    click.echo(' mikemech: v{0}'.format(__version__))
    click.echo(' Executable: {0}'.format(sys.argv[0]))
    click.echo(' Python: {0} ({1}, {2})'.format(
        platform.python_version(),
        platform.python_implementation(),
        platform.python_compiler(),
    ))


@cli.command()
@click.argument('instance', required=False)
@click.pass_context
def port(ctx, instance):
    '''Displays guest port mappings.'''

    cloud_name = ctx.obj['cloud_name']
    LOGGER.debug('cloud_name:%s instance:%s', cloud_name, instance)

    if cloud_name:
        utils.cloud_run(cloud_name, ['port'])
        return

    if platform.system() == 'Linux':
        sys.exit(click.style('This command is not supported on this OS.', fg='red'))

    if instance:
        # single instance
        instances = [instance]
    else:
        # multiple instances
        instances = utils.instances()

    # FUTURE: implement port forwarding?
    for an_instance in instances:
        inst = MechInstance(an_instance)

        if inst.provider == 'vmware':
            click.echo('Instance ({}):'. format(an_instance))
            nat_found = False
            vmrun = VMrun(inst.vmx)
            for line in vmrun.list_host_networks().split('\n'):
                network = line.split()
                if len(network) > 2 and network[2] == 'nat':
                    click.echo(vmrun.list_port_forwardings(network[1]))
                    nat_found = True
            if not nat_found:
                click.secho('Cannot find a nat network', fg='red')
        else:
            click.secho('Not yet implemented on this platform.', fg='red')


@cli.command()
@click.argument('instance', required=False)
@click.option('-s', '--show-only', is_flag=True, default=False)
@click.pass_context
def provision(ctx, instance, show_only):
    '''
    Provision the instance(s).

    Notes:

    Provision types are: 'file', 'shell' (Bash), 'ps' (Powershell), and 'pyinfra'.

    'shell' or 'ps' can be inline.

    'shell', 'ps', and 'pyinfra' can have a remote endpoint ('http', 'https', 'ftp') for the script.
    (ex: 'http://example.com/somefile.sh' or ex: 'ftp://foo.com/install.sh')

    'pyinfra' scripts must end with '.py' and 'pyinfra' must be installed.
    See https://pyinfra.readthedocs.io/en/latest/ for more info.

    Provisioning is run when the instance is started or you can re-run the provisioning.

    An example of provisioning could be installing puppet (or your config tool of choice)
    or preparing the instance "just the way you want it".

    '''

    cloud_name = ctx.obj['cloud_name']
    LOGGER.debug('cloud_name:%s show_only:%s instance:%s', cloud_name, show_only, instance)

    if cloud_name:
        utils.cloud_run(cloud_name, ['provision'])
        return

    if instance:
        # single instance
        instances = [instance]
    else:
        # multiple instances
        instances = utils.instances()

    for an_instance in instances:
        inst = MechInstance(an_instance)

        if inst.created:
            utils.provision(inst, show_only)
        else:
            click.echo('VM not created.')


@cli.command()
@click.argument('instance', required=True)
@click.pass_context
def ip(ctx, instance):
    '''
    Outputs the IP address of the instance.
    '''

    cloud_name = ctx.obj['cloud_name']
    LOGGER.debug('cloud_name:%s instance:%s', cloud_name, instance)

    if cloud_name:
        utils.cloud_run(cloud_name, ['ip', 'ip_address'])
        return

    inst = MechInstance(instance)

    if inst.created:
        ip_address = inst.get_ip()
        if ip_address:
            click.secho(ip_address, fg='green')
        else:
            click.secho('Unknown IP address', fg='red')
    else:
        click.secho('VM not created', fg='yellow')


@cli.command()
@click.argument('src', required=True)
@click.argument('dst', required=True)
@click.argument('extra-ssh-args', required=False)
@click.pass_context
def scp(ctx, src, dst, extra_ssh_args):
    '''
    Copies files to and from the instance using SCP.
    '''
    cloud_name = ctx.obj['cloud_name']
    LOGGER.debug('cloud_name:%s src:%s dst:%s extra_ssh_args:%s',
                 cloud_name, src, dst, extra_ssh_args)

    if cloud_name:
        utils.cloud_run(cloud_name, ['scp'])
        return

    dst_instance, dst_is_host, dst = dst.partition(':')
    src_instance, src_is_host, src = src.partition(':')

    instance_name = None
    if dst_is_host and src_is_host:
        sys.exit(click.style('Both src and dst are host destinations', fg='red'))
    if dst_is_host:
        instance_name = dst_instance
    else:
        dst = dst_instance
    if src_is_host:
        instance_name = src_instance
    else:
        src = src_instance

    if instance_name is None:
        sys.exit(click.style('Could not determine instance name.', fg='red'))

    inst = MechInstance(instance_name)

    if inst.created:
        _, _, stderr = utils.scp(inst, src, dst, dst_is_host, extra_ssh_args)
        if stderr != '':
            click.echo(stderr)
    else:
        click.secho('VM not created.', fg='red')


@cli.command()
@click.argument('instance', required=True)
@click.option('--command', '-c', required=False, metavar='COMMAND',
              help='Command to run on instance.')
@click.option('--plain', '-p', is_flag=True, default=False,
              help='Plain mode, leaves authentication up to user.')
@click.argument('extra-ssh-args', required=False)
@click.pass_context
def ssh(ctx, instance, command, plain, extra_ssh_args):
    '''
    Connects to an instance via SSH or runs a command (if COMMAND is supplied).
    '''
    cloud_name = ctx.obj['cloud_name']
    LOGGER.debug('cloud_name:%s instance:%s command:%s plain:%s extra_ssh_args:%s',
                 cloud_name, instance, command, plain, extra_ssh_args)

    if cloud_name:
        click.secho('Using `ssh` on cloud instance is not supported.', fg='red')
        return

    inst = MechInstance(instance)

    if inst.created:
        return_code, stdout, stderr = utils.ssh(inst, command, plain, extra_ssh_args)
        LOGGER.debug('command:%s return_code:%d stdout:%s stderr:%s',
                     command, return_code, stdout, stderr)
        if stdout:
            click.echo(stdout)
        if stderr:
            click.echo(stderr)
        sys.exit(return_code)
    else:
        click.echo('VM not created.')


@cli.command()
@click.argument('instance', required=False)
@click.pass_context
def ssh_config(ctx, instance):
    '''
    Output OpenSSH configuration to connect to the instance.
    '''
    cloud_name = ctx.obj['cloud_name']
    LOGGER.debug('cloud_name:%s instance:%s', cloud_name, instance)

    if cloud_name:
        utils.cloud_run(cloud_name, ['ssh-config'])
        return

    if instance:
        # single instance
        instances = [instance]
    else:
        # multiple instances
        instances = utils.instances()

    for an_instance in instances:
        inst = MechInstance(an_instance)
        if inst.created:
            click.echo(utils.config_ssh_string(inst.config_ssh()))
        else:
            click.secho('VM ({}) is not created.'.format(an_instance), fg='red')


@cli.command()
@click.argument('instance', required=False)
@click.pass_context
def suspend(ctx, instance):
    '''
    Suspends the instance(s).
    '''
    cloud_name = ctx.obj['cloud_name']
    LOGGER.debug('cloud_name:%s instance:%s', cloud_name, instance)

    if cloud_name:
        utils.cloud_run(cloud_name, ['suspend'])
        return

    if instance:
        # single instance
        instances = [instance]
    else:
        # multiple instances
        instances = utils.instances()

    for an_instance in instances:
        inst = MechInstance(an_instance)

        if inst.created:
            if inst.provider == 'vmware':
                vmrun = VMrun(inst.vmx)
                if vmrun.suspend() is None:
                    click.secho('Not suspended', fg='red')
                else:
                    click.secho('Suspended', fg='green')
            else:
                click.secho('Not sure equivalent command on this platform.', fg='red')
                click.secho('If you know, please open issue on github.', fg='red')
        else:
            click.secho('VM has not been created.')


@cli.command()
@click.argument('instance', required=False)
@click.option('--disable-shared-folders', is_flag=True, default=False)
@click.pass_context
def resume(ctx, instance, disable_shared_folders):
    '''
    Resume paused/suspended instance(s).
    '''
    cloud_name = ctx.obj['cloud_name']
    LOGGER.debug('cloud_name:%s instance:%s disable_shared_folders:%s',
                 cloud_name, instance, disable_shared_folders)

    if cloud_name:
        utils.cloud_run(cloud_name, ['resume', 'unpause'])
        return

    if instance:
        # single instance
        instances = [instance]
    else:
        # multiple instances
        instances = utils.instances()

    for an_instance in instances:
        inst = MechInstance(an_instance)
        inst.disable_shared_folders = disable_shared_folders
        LOGGER.debug('instance:%s', instance)

        # if we have started this instance before, try to unpause
        if inst.created:
            utils.unpause_vm(inst)
        else:
            click.secho('VM not created', fg='red')


@cli.command()
@click.argument('instance', required=False)
@click.pass_context
def upgrade(ctx, instance):
    '''
    Upgrade the VM and virtual hardware for the instance(s).

    Note: The VMs must be created and stopped.

    '''
    cloud_name = ctx.obj['cloud_name']
    LOGGER.debug('cloud_name:%s instance:%s', cloud_name, instance)

    if cloud_name:
        utils.cloud_run(cloud_name, ['upgrade'])
        return

    if instance:
        # single instance
        instances = [instance]
    else:
        # multiple instances
        instances = utils.instances()

    for an_instance in instances:
        inst = MechInstance(an_instance)

        if inst.created:
            if inst.provider == 'vmware':
                vmrun = VMrun(inst.vmx)
                state = vmrun.check_tools_state(quiet=True)
                if state == 'running':
                    click.secho('VM must be stopped before doing upgrade.')
                else:
                    if vmrun.upgradevm(quiet=False) is None:
                        click.secho('Not upgraded', fg='red')
                    else:
                        click.secho('Upgraded', fg='yellow')
            else:
                click.secho('Functionality not available on this platform.', fg='red')
        else:
            click.secho('VM ({}) not created.'.format(an_instance), fg='red')


@cli.command()
@click.argument('instance', required=False)
@click.pass_context
def pause(ctx, instance):
    '''
    Pauses the instance(s).
    '''
    cloud_name = ctx.obj['cloud_name']
    LOGGER.debug('cloud_name:%s instance:%s', cloud_name, instance)

    if cloud_name:
        utils.cloud_run(cloud_name, ['pause'])
        return

    if instance:
        # single instance
        instances = [instance]
    else:
        # multiple instances
        instances = utils.instances()

    for an_instance in instances:
        inst = MechInstance(an_instance)

        if inst.created:
            if inst.provider == 'vmware':
                vmrun = VMrun(inst.vmx)
                pause_results = vmrun.pause()
                if pause_results is None:
                    click.secho('Not paused', fg='red')
                else:
                    click.secho('Paused', fg='yellow')
            else:
                vbm = VBoxManage()
                pause_results = vbm.pause(inst.name)
                if pause_results is None:
                    click.secho('Not paused', fg='red')
                else:
                    click.secho('Paused', fg='yellow')
        else:
            click.secho('VM ({}) not created.'.format(an_instance), fg='red')


@cli.command()
@click.argument('instance', required=False)
@click.option('--force', '-f', is_flag=True, default=False, help='Force a hard stop.')
@click.pass_context
def down(ctx, instance, force):
    '''
    Stops the instance(s).
    '''
    cloud_name = ctx.obj['cloud_name']
    LOGGER.debug('cloud_name:%s instance:%s force:%s', cloud_name, instance, force)

    if cloud_name:
        utils.cloud_run(cloud_name, ['down', 'halt', 'stop'])
        return

    if instance:
        # single instance
        instances = [instance]
    else:
        # multiple instances
        instances = utils.instances()

    for an_instance in instances:
        inst = MechInstance(an_instance)

        if inst.created:
            if inst.provider == 'vmware':
                vmrun = VMrun(inst.vmx)
                if not force and vmrun.installed_tools():
                    stopped = vmrun.stop()
                else:
                    stopped = vmrun.stop(mode='hard')
                if stopped is None:
                    click.secho('Not stopped', fg='red')
                else:
                    click.secho('Stopped', fg='green')
            else:
                vbm = VBoxManage()
                stopped = vbm.stop(vmname=inst.name, quiet=True)
                if stopped is None:
                    click.secho('Not stopped', fg='red')
                else:
                    click.secho('Stopped', fg='green')
        else:
            click.secho('VM ({}) not created.'.format(an_instance), fg='red')


@cli.command()
@click.argument('instance', required=False)
@click.option('-f', '--force', is_flag=True, default=False, help='Destroy without confirmation.')
@click.pass_context
def destroy(ctx, instance, force):
    '''
    Stops and deletes all traces of the instances.
    '''
    cloud_name = ctx.obj['cloud_name']
    LOGGER.debug('cloud_name:%s instance:%s force:%s', cloud_name, instance, force)

    if cloud_name:
        utils.cloud_run(cloud_name, ['destroy'])
        return

    if instance:
        # single instance
        instances = [instance]
    else:
        # multiple instances
        instances = utils.instances()

    for an_instance in instances:
        inst = MechInstance(an_instance)

        if os.path.exists(inst.path):
            if force or utils.confirm('Are you sure you want to delete {} '
                                      'at {}'.format(inst.name, inst.path), default='n'):
                click.secho('Deleting ({})...'.format(an_instance), fg='green')

                if inst.provider == 'vmware':
                    vmrun = VMrun(inst.vmx)
                    vmrun.stop(mode='hard', quiet=True)
                    vmrun.delete_vm()
                else:
                    vbm = VBoxManage()
                    vbm.stop(vmname=inst.name, quiet=True)
                    vbm.unregister(vmname=inst.name, quiet=True)

                if os.path.exists(inst.path):
                    shutil.rmtree(inst.path)
                click.echo('Deleted')
            else:
                click.secho('Delete aborted.', fg='red')
        else:
            click.secho('VM ({}) not created.'.format(an_instance), fg='red')


@cli.command()
@click.argument('instance', required=True)
@click.pass_context
def ps(ctx, instance):
    '''
    List running processes in Guest OS.
    '''
    cloud_name = ctx.obj['cloud_name']
    LOGGER.debug('cloud_name:%s instance:%s', cloud_name, instance)

    if cloud_name:
        utils.cloud_run(cloud_name, ['ps', 'process-status', 'process_status'])
        return

    inst = MechInstance(instance)

    if inst.created:
        _, stdout, _ = utils.ssh(inst, 'ps -ef')
        click.echo(stdout)
    else:
        click.echo('VM {} not created.'.format(instance))


@cli.command()
@click.option('--purge', '-p', is_flag=True, default=False, help='Kill and remove any instances.')
@click.pass_context
def global_status(ctx, purge):
    '''
    Outputs info about all instances running on this host and optionally destroy them.
    '''
    cloud_name = ctx.obj['cloud_name']
    LOGGER.debug('cloud_name:%s purge:%s', cloud_name, purge)

    if cloud_name:
        utils.cloud_run(cloud_name, ['global-status', 'global_status', 'gs'])
        return

    if purge:
        utils.cleanup_dir_and_vms_from_dir('', all_vms=True)
    else:
        vmrun = VMrun()
        if vmrun.installed():
            click.echo('===VMware VMs===')
            click.echo(vmrun.list())

        vbm = VBoxManage()
        if vbm.installed():
            click.echo('===VirtualBox VMs===')
            click.echo(vbm.list())


@cli.command()
@click.argument('instance', required=False)
@click.option('--disable-provisioning', is_flag=True, default=False, help='Do not provision.')
@click.option('--disable-shared-folders', is_flag=True, default=False, help='Do not share folders.')
@click.option('--gui', is_flag=True, default=False, help='Start GUI, otherwise starts headless.')
@click.option('--memsize', metavar='MEMORY', help='Specify memory size in MB.')
@click.option('--no-cache', is_flag=True, default=False, help='Do not save the downloaded box.')
@click.option('--no-nat', is_flag=True, default=False,
              help='Do not use NAT networking (i.e., use bridged).')
@click.option('--numvcpus', metavar='VCPUS', help='Specify number of vcpus.')
@click.option('-r', '--remove-vagrant', is_flag=True, default=False, help='Remove vagrant user.')
@click.pass_context
def up(ctx, instance, disable_provisioning, disable_shared_folders, gui, memsize, no_cache,
       no_nat, numvcpus, remove_vagrant):
    '''
    Starts and provisions instance(s).

    Notes:

    If no instance is specified, all instances will be started.

    The options ('memsize', 'numvcpus', and 'no-nat') will only be applied
    upon first run of the 'up' command.

    The 'no-nat' option will only be applied if there is no network
    interface supplied in the box file for 'vmware'. For 'virtualbox',
    if you need internet access from the vm, then you will want to
    use 'no-nat'. Interface 'en0' will be used for bridge.

    Unless 'disable-shared-folders' is used, a default read/write
    share called 'mech' will be mounted from the current directory.
    '/mnt/hgfs/mech' on 'vmware' and '/mnt/mech' on 'virtualbox'
    To add/change shared folders, modify the Mechfile directly, then
    stop/start the VM.

    The 'remove-vagrant' option will remove the vagrant account from the
    guest VM which is what 'mech' uses to communicate with the VM.
    Be sure you can connect/admin the instance before using this option.
    Be sure to check that root cannot ssh, or change the root password.
    '''
    cloud_name = ctx.obj['cloud_name']
    LOGGER.debug('cloud_name:%s instance:%s disable_provisioning:%s disable_shared_folders:%s '
                 'gui:%s memsize:%s no_cache:%s no_nat:%s numvcpus:%s remove_vagrant:%s',
                 cloud_name, instance, disable_provisioning, disable_shared_folders,
                 gui, memsize, no_cache, no_nat, numvcpus, remove_vagrant)

    if cloud_name:
        utils.cloud_run(cloud_name, ['up', 'start'])
        return

    if instance:
        # single instance
        instances = [instance]
    else:
        # multiple instances
        instances = utils.instances()

    for an_instance in instances:
        inst = MechInstance(an_instance)

        inst.gui = gui
        inst.disable_shared_folders = disable_shared_folders
        inst.disable_provisioning = disable_provisioning
        inst.remove_vagrant = remove_vagrant
        inst.no_nat = no_nat

        if not utils.report_provider(inst.provider):
            return

        location = inst.url
        if not location:
            location = inst.box_file

        # only run init_box on first 'up'
        # extracts the VM files from the singular .box archive
        if not inst.created:
            path_to_vmx_or_vbox = utils.init_box(
                an_instance,
                box=inst.box,
                box_version=inst.box_version,
                location=location,
                instance_path=inst.path,
                save=not no_cache,
                numvcpus=numvcpus,
                memsize=memsize,
                no_nat=no_nat,
                windows=inst.windows,
                provider=inst.provider)
            if inst.provider == 'vmware':
                inst.vmx = path_to_vmx_or_vbox
            else:
                inst.vbox = path_to_vmx_or_vbox
                vbm = VBoxManage()
                if memsize:
                    vbm.memory(inst.name, memsize)
                if numvcpus:
                    vbm.cpus(inst.name, numvcpus)
                # virtualbox wants to add shared folder before starting VM
                utils.share_folders(inst)

            inst.created = True

        utils.start_vm(inst)


@cli.command()
@click.argument('name', required=True)
@click.pass_context
def remove(ctx, name):
    '''
    Remove instance from the Mechfile.
    '''
    cloud_name = ctx.obj['cloud_name']
    LOGGER.debug('cloud_name:%s name:%s', cloud_name, name)

    if cloud_name:
        utils.cloud_run(cloud_name, ['remove', 'delete', 'rm'])
        return

    mechfiles = utils.load_mechfile()
    LOGGER.debug('mechfiles:%s', mechfiles)

    inst = mechfiles.get(name, None)
    if inst:
        click.secho('Removing ({}) from the Mechfile.'.format(name), fg='green')
        utils.remove_mechfile_entry(name=name)
        click.secho('Removed from the Mechfile.', fg='green')
    else:
        sys.exit(click.style('There is no instance called '
                             '({}) in the Mechfile.'.format(name), fg='red'))


@cli.command()
@click.argument('name', required=True)
@click.argument('location', required=True)
@click.option('--add-me', '-a', is_flag=True, default=False,
              help='Add the current user/pubkey to guest.')
@click.option('--box', metavar='BOXNAME', help='Name of the box (ex: bento/ubuntu-10.04).')
@click.option('--box-version', metavar='VERSION', help='Constrain to specific box version.')
@click.option('--provider', metavar='PROVIDER', default='vmware',
              help='Provider (`vmware` or `virtualbox`)')
@click.option('--use-me', '-u', is_flag=True, default=False,
              help='Use the current user for mech interactions.')
@click.option('--windows', '-w', is_flag=True, default=False, help='Windows instance')
@click.pass_context
def add(ctx, name, location, add_me, box, box_version, provider, use_me, windows):
    '''
    Add instance to the Mechfile.

    Notes:

    The 'add-me' option will add the currently logged in user to the guest,
    add the same user to sudoers, and add the id_rsa.pub key to the authorized_hosts file
    for that user.
    '''
    cloud_name = ctx.obj['cloud_name']
    LOGGER.debug('cloud_name:%s name:%s location:%s add_me:%s box:%s box_version:%s '
                 'provider:%s use_me:%s windows:%s', cloud_name, name, location, add_me, box,
                 box_version, provider, use_me, windows)

    if cloud_name:
        utils.cloud_run(cloud_name, ['add'])
        return

    if not utils.report_provider(provider):
        return

    utils.report_provider(provider)

    click.secho('Adding ({}) to the Mechfile.'.format(name), fg='green')

    utils.add_to_mechfile(
        location=location,
        box=box,
        name=name,
        box_version=box_version,
        add_me=add_me,
        use_me=use_me,
        provider=provider,
        windows=windows)
    click.secho('Added to the Mechfile.', fg='green')


@cli.command()
@click.argument('location', required=True)
@click.option('--add-me', '-a', is_flag=True, default=False,
              help='Add the current user/pubkey to guest.')
@click.option('--box', metavar='BOXNAME', help='Name of the box (ex: bento/ubuntu-10.04).')
@click.option('--box-version', metavar='VERSION', help='Constrain to specific box version.')
@click.option('--force', '-f', is_flag=True, default=False, help='Overwrite existing Mechfile.')
@click.option('--name', metavar='INSTANCE', default='first',
              help='Name of the instance (ex: `first`).')
@click.option('--provider', metavar='PROVIDER', default='vmware',
              help='Provider (`vmware` or `virtualbox`)')
@click.option('--use-me', '-u', is_flag=True, default=False,
              help='Use the current user for mech interactions.')
@click.option('--windows', '-w', is_flag=True, default=False, help='Windows instance')
@click.pass_context
def init(ctx, location, add_me, box, box_version, force, name, provider, use_me, windows):
    '''
    Initialize Mechfile.

    Notes:

    The location can be a:

        + URL (ex: 'http://example.com/foo.box'),

        + box file (ex: 'file:/mnt/boxen/foo.box'),

        + json file (ex: 'file:/tmp/foo.json'), or

        + HashiCorp account/box (ex: 'bento/ubuntu-18.04').

    A default shared folder name 'mech' will be available
    in the guest for the current directory.

    The 'add-me' option will add the currently logged in user to the guest,
    add the same user to sudoers, and add the id_rsa.pub key to the
    authorized_hosts file for that user.

    The 'use-me' option will use the currently logged in user for
    future interactions with the guest instead of the vagrant user.

    The 'windows' flag is used for provisioning. When enabled, winrm will
    be used instead of scp.
    '''
    cloud_name = ctx.obj['cloud_name']
    LOGGER.debug('cloud_name:%s location:%s add_me:%s box:%s box_version:%s '
                 'force:%s name:%s provider:%s use_me:%s windows:%s', cloud_name,
                 location, add_me, box, box_version,
                 force, name, provider, use_me, windows)

    if not utils.valid_provider(provider):
        sys.exit(click.style('Need to provide valid provider.', fg='red'))

    if cloud_name:
        utils.cloud_run(cloud_name, ['init'])
        return

    if os.path.exists('Mechfile') and not force:
        sys.exit(click.style('`Mechfile` already exists in this directory. Remove it \n'
                             'before running `mech init` or use `mech add`.', fg='red'))

    utils.report_provider(provider)

    click.secho('Initializing mech', fg='green')
    utils.init_mechfile(
        location=location,
        box=box,
        name=name,
        box_version=box_version,
        add_me=add_me,
        use_me=use_me,
        provider=provider,
        windows=windows)
    click.secho('A `Mechfile` has been initialized and placed in this directory. \n'
                'You are now ready to `mech up` your first virtual environment!', fg='green')


MECH_ALIASES = {
    'delete': remove,
    'global-status': global_status,
    'gs': global_status,
    'halt': down,
    'ip-address': ip,
    'ip_address': ip,
    'ls': list,
    'process-status': ps,
    'process_status': ps,
    'rm': remove,
    'start': up,
    'stop': down,
    'unpause': resume,
}
